--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

   description: "Organizes all informations about identifiers";
   date: "$Date: 1996/07/03 23:04:39 $";
   revision: "$Revision: 1.1 $";
	author: "$Author: vkyr $"


class MANAGER

creation
   make

feature 

         -- Eiffel reserved words
   s_alias, s_all, s_and, s_as, s_check,
   s_class, s_creation, s_current, s_debug, s_deferred,  
   s_define, s_do, s_else, s_elseif, s_end, s_ensure,
   s_expanded, s_export, s_external, s_false,
   s_feature, s_from, s_frozen, s_if, s_implies,
   s_indexing, s_infix, s_inherit, s_inspect,
   s_invariant, s_is, s_like, s_local, s_loop,
   s_obsolete, s_old, s_once, s_or, s_prefix,
   s_redefine, s_rename, s_require, s_rescue, s_result,
   s_retry, s_select, s_separate, s_strip, s_then,
   s_true, s_undefine, s_unique, s_until, s_variant,
   s_when, s_xor: INTEGER is unique;

   is_reserved,
   is_ident,
   is_string: INTEGER is unique;

   identifier_type,
   identifier_code: INTEGER;  -- Results of find_type 

   make is
      do
			!!res_list.make  -- for SmallEiffel this is a DICTIONARY and not
         init_symbols     -- a HASH_TABLE
      end -- make      

   find_type (s: STRING) is
      require
         parameter_ok: s /= void
      do
         if res_list.has (s) then
            identifier_type := is_reserved;
				identifier_code := res_list.at (s);
         else
            identifier_type := is_ident
         end
      end -- find_type 

feature {NONE}

	  res_list:      DICTIONARY [INTEGER, STRING];
         -- store all reserved identifiers  

   dcl (a: ARRAY [INTEGER]; b: ARRAY [STRING]) is
      local
         s: STRING;
         i: INTEGER;
         ptr: INTEGER;
      do
         from
            ptr := a.lower
         until
            ptr > a.upper
         loop
            i := a.item(ptr);
            s := b.item(ptr);
            if s /= void and then i /= void then
               res_list.put (i, s);
            end;
            ptr := ptr + 1;
         end   
      end -- dcl       

   init_symbols is
      do
         dcl (<< s_alias, 
         s_all, 
         s_and, 
         s_as, 
         s_check, 
         s_class, 
         s_creation,
			s_current,
         s_debug, 
         s_deferred, 
         s_define, 
         s_do, 
         s_else, 
         s_elseif, 
         s_end, 
         s_ensure, 
         s_expanded, 
         s_export, 
         s_external, 
         s_false >>, 
         << "alias", 
         "all", 
         "and", 
         "as", 
         "check", 
         "class",
         "creation", 
			"current",
         "debug", 
         "deferred", 
         "define", 
         "do", 
         "else",  
         "elseif",
         "end",  
         "ensure", 
         "expanded", 
         "export", 
         "external", 
         "false">>);
         dcl (<< s_feature, 
         s_from, 
         s_feature, 
         s_from, 
         s_frozen, 
         s_if, 
         s_implies, 
         s_indexing, 
         s_infix, 
         s_inherit,
         s_inspect, 
         s_invariant, 
         s_is, 
         s_like, 
         s_local, 
         s_loop, 
         s_obsolete, 
         s_old, 
         s_once, 
         s_or>>,
         << "feature", 
         "from", 
         "feature", 
         "from", 
         "frozen", 
         "if", 
         "implies", 
         "indexing", 
         "infix",  
         "inherit", 
         "inspect", 
         "invariant",  
         "is", 
         "like",
         "local", 
         "loop", 
         "obsolete", 
         "old", 
         "once", 
         "or">>);
         dcl (<< s_prefix, 
         s_redefine, 
         s_rename, 
         s_require, 
         s_rescue, 
         s_result, 
         s_retry, 
         s_select, 
         s_separate, 
         s_strip, 
         s_then, 
         s_true, 
         s_undefine, 
         s_unique, 
         s_until, 
         s_variant, 
         s_when, 
         s_xor>>, 
         <<"prefix",  
         "redefine",  
         "rename",  
         "require", 
         "rescue",
         "result",
         "retry", 
         "select", 
         "separate", 
         "strip", 
         "then", 
         "true", 
         "undefine", 
         "unique", 
         "until", 
         "variant", 
         "when", 
         "xor">>);
      end; -- init_symbols         

end -- class MANAGER        
