--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

   description: "Output handler";
   date: "$Date: 1996/07/03 23:04:48 $";
   revision: "$Revision: 1.1 $";
	written_by: "$Valentino Kyriakides$"

class P_OUTPUT

creation
   make

feature

   to_lower,
   break_lines,
   line_break,
   long_list,
   upper_types,
   print_line_number: BOOLEAN

   tab_pos,
   line_number,
   lines_before,
   lines_after,
   increment_before,
   increment_after: INTEGER;

   tab_string:      STRING;
   do_tab:          INTEGER is 3;
	
	normal: INTEGER is 1;
	bold:   INTEGER is 2;
	italic: INTEGER is 3;
	bi_both: INTEGER is 4;

   fmt: FORMAT is
      once
         !!Result
      end -- fmt       

   make is
      do
         tab_string := "   "; -- 3xSPACES for *default*    		          
         long_list := true;
         line_number := 1
      end -- make      

   set_option (s: STRING) is
      do
         if s.is_equal ("-n") then
            print_line_number := true;
         elseif s.is_equal ("-x") then
            break_lines := true
         elseif s.is_equal ("-s") then
            long_list := false
         elseif s.is_equal ("-l") then
            to_lower := true
            upper_types := true
         elseif s.is_equal ("-u") then
            upper_types := true
         elseif s.is_equal ("-2") then
            tab_string := "  "
         elseif s.is_equal ("-3") then
            tab_string := "   "
         elseif s.is_equal ("-4") then
            tab_string := "    "
         elseif s.is_equal("-t") then
            tab_string := "%T";    
			elseif s.is_equal("-rtf") then
            !RTF_FORMAT!format.make;
			elseif s.is_equal("-mif") then
            !MIF_FORMAT!format.make;
			elseif s.is_equal("-mime") then
            !MIME_FORMAT!format.make;
			elseif s.is_equal("-ansi") then
            !ANSI_FORMAT!format.make;
         else
				!ASCII_FORMAT!format.make; -- default is ascii
         end
      end; -- set_option

   increment_line_number is
      do
         line_number := line_number + 1
      end -- increment_line_number

   help is
      do
         --io.set_error_default;
         io.put_string ("Usage: espretty <file> [-l] [-n] [-s] [-u] [-x] [-t|-2|-3|-4]%N");
			io.put_string ("%T%T[-ascii|-ansi|-mif|-mime|-rtf]%N%N");
         io.put_string ("Options are: %N");
			io.put_string ("%T     default are (3) blanks indent and ascii output format.%N%N");
         io.put_string ("%T-l   transforms identifiers (not Types) into lower chars.%N");
         io.put_string ("%T-n   print line numbers and statement level.%N");
         io.put_string ("%T-s   removes feature body - (.. short ..).%N");
         io.put_string ("%T-u   transforms (most) types into upper chars.%N");
         io.put_string ("%T-x   starts new line after redefine, until ...%N");
         io.put_string ("%T     removes empty lines in routine bodies.%N");
         io.put_string ("%T-t|-2|-3|-4   use TAB (8) or (n) blanks for indent.%N");
			io.put_string ("%T-ascii|    use plain ASCII for output format.%N");
			io.put_string ("%T-ansi |    use keyword boldfacing for terminal output.%N");
			io.put_string ("%T-mif  |    use Frame MIFFile output format.%N");
			io.put_string ("%T-mime |    use Mime output format.%N");
			io.put_string ("%T-rtf       use RTF output format.%N%N"); 
         io.put_string ("Examples:%N%N");
         io.put_string ("%Tspretty file.e -n >newfile -3%N");
         io.put_string ("%T   copies file.e formatted with line numbering into newfile.%N%N");
         io.put_string ("%Tspretty file.e -4 -u -l -s%N");
         io.put_string ("%T   lists a short version of file.e with upper/lower chars.%N%N");
         --io.set_output_default;
      end; -- help      

   break (before, after: INTEGER) is
      require 
         parameter_correct: before >= 0  and then after >= 0
      do
         line_break := true;
         lines_before := before;
         lines_after := after;
      end; -- break     

   tab_set (n: INTEGER) is
      do
         tab_pos := n
      end -- tab_set   

   do_increment (before, after: INTEGER) is
      do
         increment_before := before;
         increment_after := after;
      end -- do_increment         

   tab_increment (n: INTEGER) is
      do
         if n > 0 or else tab_pos > 1 then
               -- bug in structure of program handles by spretty.e      
            tab_pos := tab_pos + n
         end
      end -- tab_increment        

	outstart is
		do
			format.start_output
		end; -- outstart
		
	outend is
		do
			format.end_output
		end; -- outend
		
	comment_process (s: STRING) is
		do
			format.comment_process (s);
		end; -- comment_process
		
   sout (s: STRING; attrfmt: INTEGER) is
      require 
         s /= void
      local
         i : INTEGER;
      do
         tab_increment (increment_before);
            --
         if line_break then
               -- lines before         
            from
               i := 1  
            until
               i > lines_before
            loop
               if line_number > 1 then
                  format.next_line;
               end;
               i := i + 1;
            end;
            lines_before := 0;
               --
            if print_line_number then
               io.put_string (fmt.i2s ("%%4d", line_number));
               if tab_pos > do_tab then
						format.normal_write (":");
						format.normal_write (fmt.i2s ("%%2d", tab_pos - do_tab ));
               else
						format.normal_write("   ");
               end
            end;
               -- tabs      
            from
               i := 1 
            until
               i > tab_pos 
            loop
					format.normal_write(tab_string);
               i := i + 1;
            end      
               --
            line_break := false;
         end       
            --
         if attrfmt = bold then
				format.bold_write(s);
			elseif attrfmt = italic then
				format.italic_write(s);
			elseif attrfmt = bi_both then
				format.bi_both_write(s);
			else
				format.normal_write(s);
			end
            --
         tab_increment (increment_after);
         do_increment (0, 0);
         if break_lines and then lines_after > 0 then
            line_break := true;
            lines_before := lines_after;
            lines_after := 0;
         end;
      end -- sout      

   error_msg (s: STRING) is
      require
         s /= void
      do
         io.put_string (s);
         io.put_new_line;
      end; -- error_msg 


	format: EIFFEL_FORMAT;

end -- class P_OUTPUT       
