--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

	description: "RTF Eiffel class formatter";
	date: "$Date: 1996/07/03 23:04:52 $";
	revision: "$Revision: 1.1 $";
	modified_by: "$Valentino Kyriakides$"

class RTF_FORMAT

inherit
	
	EIFFEL_FORMAT;
		
creation 

	make
	
feature -- Public output formatters
	
	start_output is
			-- Set up output stream, used as RTF header
		once
			io.put_string ("{\rtf0\ansi\deff0{\fonttbl\f0\froman Times-Roman;}%N");
			io.put_string ("{\colortbl \red0\green0\blue0;}%N");
			io.put_string ("{\stylesheet{\fs24 \snext0 Normal;}}%N");
			io.put_string ("\paperw");
			io.put_integer (page_width * twips_per_character + 2 
				* default_margin_size);
			io.put_string ("\paperh");
			io.put_integer (default_paper_height);
			io.put_string ("\margl");
			io.put_integer (default_margin_size);
			io.put_string ("\margr");
			io.put_integer (default_margin_size);
			io.put_string ("\margt");
			io.put_integer (default_margin_size);
			io.put_string ("\margb");
			io.put_integer (default_margin_size);
			io.put_string ("\gutter0\widowctrl%N");
			io.put_string ("\sectd\f0%N");
		end; -- start_output
	
	end_output is
			-- Shut down output.	
		once
			io.put_string ("}")
		end; -- end_output
	
	end_class is
			-- Finish formatting for a class
		do
			io.put_string ("\sect%N");
		end; -- end_class
	
	hard_space is
			-- Output non-breaking space
		do
			io.put_string ("\~")
		end; -- hard_space

	hard_hyphen is
			-- Output non-breaking hyphen
		do
			io.put_string ("\_")
		end
	
	space is
			-- Output space
		do
			io.put_string (" ")
		end; -- space
		
	write,
	output (s: STRING) is
			-- Post process output strings and determine also used 
			-- special chars for comment processing.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > s.count
			loop
				inspect
					s.item (i)
				when '{', '}', '\' then
					io.put_character ('\');
					io.put_character (s.item (i))
				when ' ' then
					hard_space
				when '-' then
					hard_hyphen
				when '`'	then		
					if is_comment then  -- Inside a comment?
						italic_on
					else
						io.put_character (s.item (i))
					end
				when right_quote then
					if is_comment then  -- Inside a comment?
						italic_off
					else
						io.put_character (s.item (i))
					end
				else
					io.put_character (s.item (i))
				end;
				i := i + 1
			end
		end; -- write, output
		
	bold_on is
			-- Output Bold start	
		do
			io.put_string ("\b ")
		end; -- bold_on
	
	bold_off is
			-- Output Bold finish	
		do
			io.put_string ("\b0 ")
		end; -- bold_off
	
	italic_on is
			-- Output Italic start	
		do
			io.put_string ("\i ")
		end; -- italic_on
	
	italic_off is
			-- Output Italic end	
		do
			io.put_string ("\i0 ")
		end; -- italic_off

	normal_on is
			-- Output Normal start	
		do
		end; -- normal_on
	
	normal_off is
			-- Output Normal end	
		do
		end; -- normal_off

	next_line is
			-- Output a linefeed
		do
			io.put_string ("\par %N")
		end; -- next_line
		
feature {NONE} -- Size constants
		
	default_paper_width: INTEGER is 11908;
	
	default_paper_height: INTEGER is 16834;
	
	default_margin_size: INTEGER is 1434;
	
	twips_per_inch: INTEGER is 1440;

	indent_width: INTEGER is 80;

	page_width: INTEGER is 80;
	
	
	twips_per_character: INTEGER is 113
		--once
		--	Result := (default_paper_width - 2 
		--		* default_margin_size) / 80
		--end; -- twips_per_character
	
	Twips_per_indent: INTEGER is 90;


feature {NONE} -- Comment processing.
	
	comment_process (s: STRING) is
			-- Checks/handles in comments the "-- class name" and the
			-- "-- featurename" matching. Both the word "class" and the
			-- "featurename" must start at the 4th charcter position inside 
			-- the comment, otherwise they will not be matched.
			-- NOTE: This is a quick'n'dirty, not well working hack, which
			-- only uses simple character matching in comment strings.
		do
			is_comment := true
		--	if s.substring_index("class",1) = 4 then 
		--		s.right_adjust
		--		s.replace_substring_all("class ","class `");
		--		s.append("'");
		--	else
		--		s.right_adjust
		--		if s.occurrences(' ') = 1 then 
		--			s.replace_substring_all("-- ","-- `");
		--			s.append("'");
		--		end
		--	end
		end; -- comment_process
		
	left_quote: CHARACTER is '`';
	right_quote: CHARACTER is '%'';

	is_comment: BOOLEAN
	
end -- class RTF_FORMAT
