--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

   description: "The scanner";
   date: "$Date: 1996/07/03 23:05:26 $";
   revision: "$Revision: 1.1 $"

class SCANNER

   inherit
      S_SUPPLY  

creation make

feature {ANY}

   make (s: F_INPUT; l: INTEGER) is
      require
         no_input_file: s /= void
         lookahead : l >= 1
      local
         i: INTEGER
      do
         input_stream := s
         input_stream.get_char
         !!type_list.make (1, l)
         !!token_list.make (1, l)
         put_ptr := 1 
         get_ptr := 0 
         from
            i := 1
         until
            i > l - 1
         loop
            next 
            into_queue
            i := i+1
         end      
      end -- make      

   last_token: STRING is
      do
         Result := token_list.item (get_ptr)
      end -- last_token

   last_token_type: INTEGER is
      do
         Result :=  type_list.item (get_ptr)
      end -- last_token_type      

   lookahead_type (i: INTEGER): INTEGER is
      require
         lookahead_negative: i < 0
         wrong_lookahead: i <= type_list.count
      local
         ptr: INTEGER
      do
         ptr := get_ptr + i
         if ptr > type_list.count then
            ptr := ptr - type_list.count
         end        
         Result := type_list.item (ptr)
      end -- lookahead_type       

   lookahead_token (i: INTEGER): STRING is
      require
         lookahead_negative: i < 0
         wrong_lookahead: i <= type_list.count
      local
         ptr: INTEGER
      do
         ptr := get_ptr + i
         if ptr > token_list.count then
            ptr := ptr - token_list.count
            Result := token_list.item (ptr)
         end       
      end -- lookahead_token      

   get_token is
      do
         next
         into_queue
         get_ptr := get_ptr + 1
         if get_ptr > type_list.count then
            get_ptr := 1
         end        
      end -- get_token 

   end_of_input: BOOLEAN is
      do
         Result := last_token_type = tc_eof
      end -- end_of_input         

   type_list:  ARRAY [INTEGER]

feature {NONE}

   input_stream: F_INPUT
   token_list: ARRAY [STRING]
   put_ptr,
   get_ptr:    INTEGER
   act_token:  TOKEN

   next is
      do
         if input_stream.end_of_input then
            !T_EOF!act_token
         else
            inspect 
               char_type (input_stream.last_char)
            when ch_letter then   !T_IDENT!act_token
            when  ch_digit then   !T_NUMBER!act_token
            when  ch_quote then   !T_STRING!act_token   
            when  ch_space then   !T_SPACE!act_token
            when  ch_newline then !T_LINE!act_token
            else                  !T_SYMB!act_token
            end   
         end        
         act_token.init (input_stream)
         act_token.get_token
         act_token.strip_text
         act_token.action 
      end -- next      

   into_queue is
      do
         type_list.put (act_token.last_token_type, put_ptr)
         token_list.put (act_token.last_token, put_ptr)
         put_ptr := put_ptr + 1
         if put_ptr > type_list.count then
            put_ptr := 1
         end       
      end -- into_queue           

end -- class scanner        
