--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------


indexing

   description: "Eiffel source pretty printer version for SmallEiffel";
   date: "$Date: 1996/07/03 22:58:08 $";
   revision: "$Revision: 1.1 $";
	written_by: "$Valentino Kyriakides$"

class ESPRETTY

   inherit
      STD_FILE
      S_SUPPLY
      MANAGER

creation main

feature 

   class_pos: INTEGER is 0;
   inherit_pos: INTEGER is 1;
   define_pos: INTEGER is 3;
   feature_pos: INTEGER is 0;
   do_pos: INTEGER is 2;
   comment_pos: INTEGER is 1; 
   break_before_feature: INTEGER is 2;

   sc : SCANNER;
   in : F_INPUT;
   m  : MANAGER;
   p  : P_OUTPUT;

   empty_line,
   level,
   brackets,
   state: INTEGER

   and_or,
   next_upper, 
   def_class: BOOLEAN

   st_class,
   st_creation,
   st_require,
   st_inherit,
   st_deferred,
   st_feature_body,
   st_feature_header, 
   st_feature_end,
   st_class_end: INTEGER is unique;
	
	normal: INTEGER is 1
	bold:   INTEGER is 2
	italic: INTEGER is 3
	bi_both: INTEGER is 4; 

   main is
      local
         i : INTEGER
      do
         !!p.make   -- Create output handler
            -- Check/setup if there are arguments/options.
         if argument_count = 0 then
            p.help
         else
            from
               i := 1
            until
               i > argument_count
            loop
               p.set_option (argument (i));
               i := i + 1
            end
            if p.long_list then 
               empty_line := 2
            else 
               empty_line := 1;
            end  
         end
            -- Check file access 
         if argument_count > 0 then
				if file_exists(argument (1)) then
				   !F_INPUT!in.access (argument (1)) 
               in.open
               !!sc.make (in, 1)		-- Create scanner
               !!m.make					-- Create symbol manager
					p.outstart        	-- Process input
               action
					p.outend
                  -- put a final %N at the end of output       
               io.put_string("%N");
            else
                  -- Wrong filename or file not accessible exit app.
               io.put_string(argument (0)); io.put_string (":  No such file '");
               io.put_string(argument (1)); io.put_string ("' %N");
            end
         end
      end -- main      

   action is
      do
         from
            sc.get_token
         until
            sc.end_of_input
         loop
            output_symbol := true
				--
            inspect 
               sc.last_token_type
            when tc_line then
               p.increment_line_number;
               if p.line_break then
                  if state /= st_feature_body or else not p.break_lines then
                        -- "manual" line breaks only outside of feature body     
                        -- if -x option was set 
                     p.break (empty_line, 0);
                  end
               else
                  p.break (1, 0);
               end       
               output_symbol := false;
            when tc_comment then						                  
               if state = st_feature_end then
                  p.tab_set (comment_pos);
               elseif state = st_feature_header then
                  p.do_increment (2, -2)
               elseif state = st_feature_body then
                  p.do_increment (1, -1)
               end
					p.comment_process (sc.last_token); -- Comment handling
            when tc_ident then
               identifier
               output_symbol := false;
            when tc_space then
               if p.line_break then
                     -- No spaces directly after line break        
                  output_symbol := false;
               end;
            when tc_symb then
               and_or := false;
               if sc.last_token.count = 1 then
                  inspect sc.last_token.item (1)
                  when ':' then
                     if state /= st_require then
                        next_upper := true
                     end
                  when '[' , '{' then
                     brackets := brackets + 1
                  when ']', '}' then
                     brackets := brackets - 1
                  else
                  end
               end        
            else
                  -- No extra commands    
            end -- when      
            --
            if print_allowed and then output_symbol then 
               p.sout (sc.last_token, normal)
            end        
            sc.get_token
         end   
      end -- action    

   identifier is
      local
         t: STRING;
      do
         t := clone (sc.last_token);
         t.to_lower;
         m.find_type (t);
         if p.to_lower then
            sc.last_token.to_lower;
         end
         --
         if m.identifier_type = m.is_reserved then
            special_action (m.identifier_code)
         else
            and_or := false;
            if state = st_feature_end then
               state := st_feature_header
               p.tab_set (feature_pos + 1); 
               p.do_increment (0, 0);
               p.break (break_before_feature, 0);
               level := 0;
            end;
            if print_allowed then
               if p.upper_types and then  
                  (next_upper or (brackets > 0) or 
                  (state = st_inherit and p.tab_pos = inherit_pos + 1)) then               
                  sc.last_token.to_upper
                  next_upper := false;
               end
               if output_symbol then 
                  p.sout (sc.last_token, italic)
               end
            end
         end
      end -- identifier

   special_action (n: INTEGER) is
      do
         inspect n 
         when s_end then
            special_end;
         when s_class then
            next_upper := true;
            if state /= st_deferred then
               def_class := false;
               p.tab_set (class_pos);
               p.break (2, 0)
            end;
            state := st_class;
         when s_feature, s_invariant then
            state := st_feature_header;
            p.break (2, 0);
            p.tab_set (feature_pos);
            p.do_increment (0, 1); 
         when s_inherit then
            state := st_inherit;
            p.break (2, 1);
            p.tab_set (inherit_pos);
            p.do_increment (0, 1);
         when s_indexing then
            p.break (2, 1);
            p.tab_set (class_pos);
            p.do_increment (0, 1);
         when s_define, s_redefine, s_select, 
            s_undefine, s_export, s_rename then
            p.break (1,1);
            p.tab_set (inherit_pos + 2);
            p.do_increment (0, 1);
         when s_creation then
            state := st_creation;
            p.break (2, 0);
            p.tab_set (feature_pos);
            p.do_increment (0, 1);
         when s_do, s_external, s_once, s_local, s_ensure, s_rescue,
			     s_alias, s_obsolete then
            level := 1;
            p.break (1, 0);
            p.tab_set (do_pos);
            p.do_increment (0,  1);
            state := st_feature_body;
         when s_require then
            state := st_require;
            level := 1;
            p.break (1, 0);
            p.tab_set (do_pos);
            p.do_increment (0,  1);
         when s_deferred then
            if def_class = false then
               def_class := true;
               p.break (2, 0);
               p.tab_set (class_pos);
               state := st_deferred
            else
               level := 1;
               p.break (1, 0); 
               p.tab_set (do_pos);
               p.do_increment (0, 1);  
               state := st_feature_body;
            end    
         when s_if , s_debug then
            p.break (1, 0);
            p.do_increment (0, 1);
            level := level + 1;
         when s_from then
            p.break (1, 1);
            p.do_increment (0, 1);
            level := level + 1;
         when s_elseif, s_loop then
            p.do_increment (-1, 1);
            p.break (1, 0);
         when s_else then
            if and_or then
               and_or := false
            else
               p.do_increment (-1, 1);
               p.break (1, 0)
            end
         when s_until then
            p.do_increment (-1, 1);
            p.break (1, 1);
         when s_inspect then
            level := level + 1;
            p.do_increment (0, 1);
            p.break (1, 0);
         when s_when then
            p.do_increment (-1, 1);
            p.break (1, 0);
         when s_and, s_or then
            and_or := true;
         when s_then  then
            and_or := false;
         else
            p.do_increment (0, 0);
               -- do nothing
         end -- inspect   
         --
         if print_allowed then
            if n = s_result then
               p.sout ("Result",italic)
				elseif n = s_current then
					p.sout ("Current",italic)
            else
               p.sout (sc.last_token, bold); 
            end;
         end
      end -- special_action       

   special_end is
      do
         if (state = st_feature_end) or (state = st_feature_header) then
            state := st_class_end;
               -- end of class         
            p.tab_set (feature_pos);
            p.do_increment (-1, 0); 
            p.break (2, 0); 
         else
            if state = st_feature_body then
               p.do_increment (-1, 0)
               level := level - 1;
               if level < 0 then
                  p.error_msg ("-- too many closing ends.");
                     -- program bug ..       
                  level := 0 
               end
               --
               if level = 0 then
                     -- end of feature declaration      
                  state := st_feature_end;
               end
            elseif state = st_inherit then
               p.tab_set (inherit_pos + 2);
               p.do_increment (0, -1);
            else
               p.do_increment (-1, -1)
            end
            p.break (1, 0);
         end
      end -- special_end          

feature {NONE}

   output_symbol: BOOLEAN;
	
   print_allowed: BOOLEAN is
      do
         Result := p.long_list or else 
         (state /= st_feature_body and state /= st_feature_end)
      end -- print_allowed        

end -- class ESPRETTY        
