--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

	description: "HTML Eiffel class formatter";
	date: "$Date: 1996/07/22 14:36:57 $";
	revision: "$Revision: 1.3 $";
	written_by: "$Valentino Kyriakides$"

class HTML_FORMAT

inherit
	
	EIFFEL_FORMAT;
		
creation 

	make
	
feature -- Public output formatters
	
	start_output is
			-- Set up output stream, used as HTML header
		once
			io.put_string ("<HTML>%N");
			io.put_string ("<HEAD>%N");
			io.put_string ("<TITLE>Produced by espretty for SmallEiffel</TITLE>%N");
			io.put_string ("</HEAD>%N");
			io.put_string ("<BODY>%N%N");
			io.put_string ("<CODE>%N%N");
		end; -- start_output
	
	end_output is
			-- Shut down output.	
		once
			io.put_string ("</CODE>%N%N");
			io.put_string ("</BODY>%N")
			io.put_string ("</HTML>%N")
		end; -- end_output
	
	end_class is
			-- Finish formatting for a class
		do
			io.put_string ("%N");
		end; -- end_class
	
	hard_space is
			-- Output non-breaking space.
			-- This is for correct space indents in html mode.
		do
			io.put_string ("&nbsp;")
		end; -- hard_space

	hard_hyphen is
			-- Output non-breaking hyphen.
			-- As an alternative we might use here
			-- also io.put_string ("&shy;").
		do
			io.put_string ("-")
		end
	
	space is
			-- Output space
		do
			io.put_string (" ")
		end; -- space
		
	write,
	output (s: STRING) is
			-- Post process output strings and determine also used 
			-- special chars for comment processing.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > s.count
			loop
				inspect
					s.item (i)
				when '{', '}', '\' then
					io.put_character (s.item (i))
				when ' ' then
					hard_space
				when '-' then
					hard_hyphen
				when '`'	then		
					if is_comment then  -- Inside a comment?
						italic_on
					else
						io.put_character (s.item (i))
					end
				when right_quote then
					if is_comment then  -- Inside a comment?
						italic_off
					else
						io.put_character (s.item (i))
					end
				else
					io.put_character (s.item (i))
				end;
				i := i + 1
			end
		end; -- write, output
		
	bold_on is
			-- Output Bold start	
		do
			io.put_string ("<B>")
		end; -- bold_on
	
	bold_off is
			-- Output Bold finish	
		do
			io.put_string ("</B>")
		end; -- bold_off
	
	italic_on is
			-- Output Italic start	
		do
			io.put_string ("<I>")
		end; -- italic_on
	
	italic_off is
			-- Output Italic end	
		do
			io.put_string ("</I>")
		end; -- italic_off

	normal_on is
			-- Output Normal start	
		do
		end; -- normal_on
	
	normal_off is
			-- Output Normal end	
		do
		end; -- normal_off

	next_line is
			-- Output a linefeed
		do
			is_comment := false   -- VKYR: important for correct comment handling
			io.put_string ("<BR>%N")
		end; -- next_line
		

feature {NONE} -- Comment processing.
	
	comment_process (s: STRING) is
			-- Checks/handles in comments the "-- class name" and the
			-- "-- featurename" matching. Both the word "class" and the
			-- "featurename" must start at the 4th charcter position inside 
			-- the comment, otherwise they will not be matched.
			-- NOTE: This is a quick'n'dirty, not well working hack, which
			-- only uses simple character matching in comment strings.
		do
			is_comment := true   -- VKYR: important for correct comment handling
			
		--	if s.substring_index("class",1) = 4 then 
		--		s.right_adjust
		--		s.replace_substring_all("class ","class `");
		--		s.append("'");
		--	else
		--		s.right_adjust
		--		if s.occurrences(' ') = 1 then 
		--			s.replace_substring_all("-- ","-- `");
		--			s.append("'");
		--		end
		--	end
		end; -- comment_process
		
	left_quote: CHARACTER is '`';
	right_quote: CHARACTER is '%'';

	is_comment: BOOLEAN
	
end -- class HTML_FORMAT
