--|----------------------------------------------------------------------------
--|      Part of the [E]iffel [s]ource [pretty] printer for SmallEiffel
--|                       
--|                    GNU GENERAL PUBLIC LICENSE
--|
--|    This program is free software; you can redistribute it and/or modify
--|    it under the terms of the GNU General Public License as published by
--|    the Free Software Foundation!
--|----------------------------------------------------------------------------

indexing

	description: "FrameMaker MIF Eiffel class formatter";
	date: "$Date: 1996/07/22 14:37:15 $";
	revision: "$Revision: 1.3 $";
	author: "$Valentino Kyriakides Kyriakides$"

class MIF_FORMAT

inherit
	
	EIFFEL_FORMAT;
		
creation 

	make
	
feature -- Public output formatters
	
	start_output is
			-- Set up output stream, used as MIF header
		once
			io.put_string ("<MIFFile 3.00> # Generated by spretty v1.1%N%N");
			io.put_string ("<Para%N");
			io.put_string ("  <PgfTag `EProgList'>%N");
			start_paraline;
		end; -- start_output
	
	end_output is
			-- Shut down output.	
		once
			end_paraline;
			io.put_string ("> # end of Para%N%N")
			io.put_string ("# End of MIFFile")
		end; -- end_output
	
	end_class is
		do
		end;

	start_string is
			-- Start a string
		do
			io.put_string ("    <String `");
			is_string := true
		end; -- start_string
			
	end_string is
			-- End a string
		do
			io.put_string ("'># end%N");
			is_string := false
		end; -- end_string
	
	hard_space is
			-- Output non-breaking space
		do
			if is_string then
				end_string;
				io.put_string ("    <Char HardSpace >%N")
				start_string;
			else
				io.put_string ("    <Char HardSpace >%N")
			end
		end; -- hard_space


	start_paraline is
			-- Start of paraline
		do
			io.put_string ("  <ParaLine%N")
			is_paraline := true;
		end -- start_paraline

	end_paraline is
			-- End of paraline
		do
			io.put_string ("  > # end of ParaLine%N")
			is_paraline := false;
		end -- end_paraline
	
	space is
			-- Output a space	
		do
			io.put_string (" ")
		end; -- space
		
	write,
	output (s: STRING) is
			-- Post process output strings and determine also used 
			-- special chars for comment processing.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > s.count
			loop
				inspect
					s.item (i)
				when '>', '\', tabstop then
					io.put_character ('\');
					io.put_character (s.item (i))
				when ' ' then
					hard_space
				when '`'	then		
					if is_comment then  -- Inside a comment?
						end_string;
						italic_on;
					else
						io.put_string ("\xd4 ")
					end
				when right_quote then
					if is_comment then  -- Inside a comment?
						italic_off;
						normal_on;
						normal_off;
						start_string;
					else
						io.put_string ("\xd5 ")
					end
				else
					io.put_character (s.item (i))
				end;
				i := i + 1
			end
		end; -- write, output
		
	bold_on is
			-- Output Bold start	
		do
			io.put_string ("    <Font %N") 
			io.put_string ("      <FTag `Bold'>%N")
			io.put_string ("      <FFamily `Times'>%N")
			io.put_string ("      <FVar `Regular'>%N")
			io.put_string ("      <FWeight `Bold'>%N")
			io.put_string ("      <FAngle `Regular'>%N")
			io.put_string ("      <FSize  12.0 pt>%N")
			io.put_string ("      <FPairKern Yes >%N")
			io.put_string ("    > # end of Font%N")
			io.put_string ("    <String `")
			is_string := true
		end; -- bold_on
	
	bold_off is
			-- Output Bold finish	
		do
			io.put_string ("'># end%N")
			is_string := false
		end; -- bold_off
	
	italic_on is
			-- Output Italic start	
		do
			io.put_string ("    <Font %N") 
			io.put_string ("      <FTag `Emphasis'>%N")
			io.put_string ("      <FFamily `Times'>%N")
			io.put_string ("      <FVar `Regular'>%N")
			io.put_string ("      <FWeight `Regular'>%N")
			io.put_string ("      <FAngle `Italic'>%N")
			io.put_string ("      <FSize  12.0 pt>%N")
			io.put_string ("      <FPairKern Yes >%N")
			io.put_string ("    > # end of Font%N")
			io.put_string ("    <String `")
			is_string := true
		end; -- italic_on
	
	italic_off is
			-- Output Italic end	
		do
			io.put_string ("'># end%N")
			is_string := false
		end; -- italic_off

	normal_on is
			-- Output Normal start	
		do
			io.put_string ("    <Font %N") 
			io.put_string ("      <FTag `'>%N")
			io.put_string ("      <FFamily `Times'>%N")
			io.put_string ("      <FVar `Regular'>%N")
			io.put_string ("      <FWeight `Regular'>%N")
			io.put_string ("      <FAngle `Regular'>%N")
			io.put_string ("      <FSize  12.0 pt>%N")
			io.put_string ("      <FPairKern Yes >%N")
			io.put_string ("    > # end of Font%N")
			io.put_string ("    <String `")
			is_string := true
		end; -- normal_on	

	normal_off is
			-- Output Normal end	
		do
			io.put_string ("'># end%N")
			is_string := false
		end; -- normal_off

	next_line is
			-- Output a linefeed
		do
		   is_comment := false   -- VKYR: important for correct comment handling
			if is_paraline then
				io.put_string ("    <Char HardReturn >%N")
				end_paraline;
				start_paraline;
			else
				start_paraline;
				io.put_string ("    <Char HardReturn >%N")
				end_paraline;
			end
		end; -- next_line
		

feature {NONE} -- Comment processing.
	
	comment_process (s: STRING) is
			-- Checks/handles in comments the "-- class name" and the
			-- "-- featurename" matching. Both the word "class" and the
			-- "featurename" must start at the 4th charcter position inside 
			-- the comment, otherwise they will not be matched.
			-- NOTE: This is a quick'n'dirty, not well working hack, which
			-- only uses simple character matching in comment strings.
		do
			is_comment := true   -- VKYR: important for correct comment handling
			
			-- if s.substring_index("class",1) = 4 then
			--	   s.right_adjust
			--	   s.replace_substring_all("class ","class `");
			--	   s.append("'");
		   --	else
			--	   s.right_adjust
			--	   if s.occurrences(' ') = 1 then
			--		   s.replace_substring_all("-- ","-- `");
			--		   s.append("'");
			--	   end
		  --	end
		end; -- comment_process
		
	left_quote: CHARACTER is '`';
	right_quote: CHARACTER is '%'';
	tabstop: CHARACTER is '%T';

	is_comment: BOOLEAN
	is_paraline: BOOLEAN
	is_string: BOOLEAN
	
end -- class MIF_FORMAT
