------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--    S Y S T E M . F I N A L I Z A T I O N _ I M P L E M E N T A T I O N   --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                            $Revision: 1.13 $                              --
--                                                                          --
--           Copyright (c) 1992,1993,1994 NYU, All Rights Reserved          --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. --
--                                                                          --
------------------------------------------------------------------------------

--  ??? this package should be a private package. It is set as public for now
--  in order to simplify testing

package System.Finalization_Implementation is

   type Root;

   subtype Finalizable     is            Root'Class;
   type    Finalizable_Ptr is access all Root'Class;

   type Root is abstract tagged record
      Previous : Finalizable_Ptr;
      Next     : Finalizable_Ptr;
   end record;

   procedure Initialize (Object : in out Root) is abstract;
   procedure Finalize   (Object : in out Root) is abstract;


   -------------------------------------------------
   --  Finalization Management Abstract Interface --
   -------------------------------------------------

   Global_Final_List : Finalizable_Ptr;
   --  This list stores the controlled objects defined in library-level
   --  packages. They will be finalized after the main program completion.

   procedure Finalize_Global_List;
   --  The procedure to be called in order to finalize the global list;

   procedure Attach_To_Final_List
    (L   : in out Finalizable_Ptr;
     Obj : in out Finalizable);
   --  Put the finalizable object on a list of finalizable elements. This
   --  procedure is called during the initialization of the controlled object.

   procedure Finalize_List (L : Finalizable_Ptr);
   --  Call Finalize on each element of the list L;

   procedure Finalize_One
    (From : in out Finalizable_Ptr;
     Obj  : in out Finalizable);
   --  Call Finalize on Obj and remove it from the list From.

end System.Finalization_Implementation;


----------------------
-- REVISION HISTORY --
----------------------

--  ----------------------------
--  revision 1.11
--  date: Tue Sep 13 09:47:20 1994;  author: comar
--  Define a fake primitive operation in order to align disaptch tables for
--  limited and not limited root controlled types. Solve the mysterious bug
--  lurking in the DOS version (Allocate was the next primitive op. and was
--  called instead of Finalize)
--  ----------------------------
--  revision 1.12
--  date: Tue Sep 13 16:40:29 1994;  author: dewar
--  Add ??? comment to Faked_Eq.
--  ----------------------------
--  revision 1.13
--  date: Thu Sep 15 14:52:55 1994;  author: comar
--  Simplify the spec tremendously. In particular, get rid of all the junk
--   remaining from RM4.0
--  ----------------------------
--  New changes after this line.  Each line starts with: "--  "
