"======================================================================
|
|   An early attempt at the C declaration parser.
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:54:50 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
|
"

FileStream require: #PreprcessorStream -> 'CPP.st' !
FileStream require: #CSymbol -> 'CSymbol.st'!

"Here's the probelm:

at what level is the symbol table management done?  we could do it here,
just parsing whatever we come to, and recording the declarations into the
symbol table.  

Alternatively (the way the code is set up), this code just returns the
 declarations in some form (symbol table entries would be ok), and allows
the higher level code to make the determinatino about how to deal with them.
 This allows for some degree of reentrancy to the system.

The symbol table itself is available to both levels, and is kept as up to date
 as possible.

I like the reentrancy aspect enough more than the pain of having to hang onto
 some partially formed declaration, so that's what we'll go with.

"


Object subclass: #CTypeParser
       instanceVariableNames: 'stream'
       classVariableNames: ''
       poolDictionaries: 'CKeywords'
       category: nil
!


CTypeParser comment:
'I parse C declarations (and some simple definitions), and yield useful
structures.  I appear to my caller to be a stream of declaration objects,
which can have "declare" invoked on them to make them register themselves
with the symbol table.  I return nil if I cannot parse, including eof.'
!

!CTypeParser class methodsFor: 'instance creation'!

on: aStream
    ^self new init: aStream
! !


!CTypeParser methodsFor: 'accessing'!

next
    "Yields nil at end of stream, otherwise, yields the next
     declaration"
    stream atEnd
	ifTrue: [ ^ nil ].
    
    ^self parseBaseType
! !



!CTypeParser methodsFor: 'parsing'!

parseStruct
    self parseStructOrUnion: 
	[ :structName :members | ^CStructDecl name: structName
					      members: members ]
!

parseUnion
    self parseStructOrUnion: 
	[ :structName :members | ^CUnionDecl name: structName
					     members: members ]
!

parseStructOrUnion: parentBlock
    "assumes we're at 'struct/\ ...' "
    | token structName members |
    token _ stream next.
    token class == IdentifierTok
	ifTrue: [ structName _ token. ].
    stream next.		"could check for it being '{'"
    members _ OrderedCollection new.
    [ token _ stream peek. 
      token class == CloseBraceTok ]
	whileFalse: 
	    [ members addAllLast: self parseBitfieldMember
	     ].
    stream next.		"gobble '}'"
    parentBlock value: structName value: members
!
	
parseBitfieldMember
    | type token names variable fieldSize |
    "parses /\<type> [ { <name>[ ':' <size> ] # ',' } ] ';'"
    "structure member are mostly like normal variable declarations, that is
     they have a type, a name (optional), and a  size (optional).  The name
     includes the other things like pointer, array, function, etc.  There
     is a comma separated list of these, terminated by a semicolon."
    "Returns an ordered collection of members; each member corresponding
     to a single declaration (typically only 1).  If there are several 
     (in the comma list case), the collection has one per item; they
     are treated as if they had all occurred on separate lines, because
     we don't care about knowing how they were grouped in the source."
    type _ self parseBaseType.
    names _ OrderedCollection new.
    [ token _ stream peek.
      token class == SemiTok ] whileFalse:
	  [ token class == CommaTok "hack-o-rama"
		ifTrue: [ stream next.  "eat the comma"
			  token _ stream peek ].
	    variable _ self parseVarDecl. "may return nil"
	    token _ stream peek.
	    token class == ColonTok
		ifTrue: [ stream next. "gobble it"
			  fieldSize _ stream next value. ].
	    "!!! need to build these guys into an entity, and add it
	     to the names list"
	    ].
    stream next.		"gobble ';'"
    ^CStructMemberDecl name: variable name type: variable type 
		       bits: fieldSize
!      
    
parseBaseType
    "Makes assumptions that all required types are defined by the point
     in time that this is encountered.  Uses type symbol table information
     to help control the parsing."
    "more interesting grammar:
     <decl> ::= <opt storage class> <opt dynamic properties>
                <type list> <var decl> ;
     <opt storage class> ::= 
                         |   static 
			 | auto 
			 | register 
			 | extern  ;

     <opt dynamic properites > ::= 
     			       | const 
     			       | volatile ;

     <type list> ::= { <identifer> }+ ;
     <var decl> ::= [ <ident> ] <suffix list>
                | '*' <var decl> 
                | '(' <var decl> ')' <suffix list>
     <suffix list> ::= 
                   | '[' <expression_const> ']' <suffix_list>
                   | '(' <param decls> ')' <suffix_list>
     "

    "can be [ static | auto | register | extern ] [ const | volatile ] 
          <list of types> <vardecl>
     <var decl> ::= [ <ident> ] <suffix list>
                | '*' <var decl> 
                | '(' <var decl> ')' <suffix list>
     
     <suffix list> ::= 
                   | '[' <expression_const> ']' <suffix_list>
                   | '(' <param decls> ')' <suffix_list>
     "
     " !!! not done "

    | token |
    token _ stream peek.
    (token == StaticKey) | (token == AutoKey) | (token == RegisterKey)
	| (token == ExternKey) 
	ifTrue: 
	    [ stream next.
	      "!!! do something"
	      token _ stream peek. ].
				 
    (token == ConstKey) | (token == VolatileKey) 
	ifTrue: [ 
		  stream next. "!!! do something"
		  token _ stream peek].
!	
    
"
really, need to parse a base type declaration, and then get the name
out of it.  The same paring should handle both top level variable decls

 	int foo;

 	char *fred;

and struct member declarations.

there are several ways to do this:
1) parse identifiers until you reach a non-identifier.  If you hit
   one of them, you are in 
   a) it's a terminator (; or ,).  The thing immediately before is
      the variable (although not in structures, necessarily)
   b) its a *.  You have a pointer of some sort, and the rest
      of the stuff you parse will not be type names.
   c) its a (.  Some kind of nested declaration.  You are past the 
      top level type names.
2) parse identifiers, until you come to something that's either not
   an identifier, or not one which is known to be a type.

"


parseVarDecl
    | token suffixes subVarDecl |
    "!!! note that here the identifier may be missing (as in casts or
     parameter declarations"
    token _ stream peek.
    (token class == COperatorToken
	and: [ token value = '*' ])
	ifTrue: [ stream next.
		  subVarDecl _ self parseVarDecl.
		  ^3 "!!! build a var decl with the * on it and return"
		  ].
    token class == OpenParenTok 
	ifTrue: [ stream next.	"gobble '('"
		  subVarDecl _ self parseVarDecl.
		  stream next.
		  suffixes _ self parseSuffixList.
		  ^3 "!!! build var decl and leave" ].
    token class == IdentifierTok
	ifTrue: [ stream next.	"gobble identifier"
		  suffixes _ self parseSuffixList.
		  ^3 "!!!something more" ].
    ^nil			"couldn't parse any more"
!


parseSuffixList
    | token remainingSuffixes arraySize params | 
    token _ stream peek.
    token class == OpenBracketTok
	ifTrue: [ stream next.	"gobble '['"
		  arraySize _ self parseExpr.
		  "### could check for ]"
		  stream next.	"gobble ']'"
		  remainingSuffixes _ self parseSuffixList.
		  ^3 "!!! combine them and return" ].
    token class == OpenParenTok
	ifTrue: [ "have a function declaraction"
		  stream next.	"gobble '('"
		  params _ self parseParamDecls.
		  "### could check for )"
		  stream next.	"gobble ')'"
		  remainingSuffixes _ self parseSuffixList.
		  ^3 "!!! combine them and return" ].
    ^nil			"couldn't parse any more"
!
		  
		  
parseParamDecls
    "expecting
         '(' /\{ <paramDecl> # ',' }* ')' "
    | token decls baseType varDecl |
    decls _ OrderedCollection new.
    [ token _ stream peek.
      token class == CloseParenTok ]
	whileFalse: 
	    [ (token class == CBinaryOperatorToken 
		   and: [ token value = ',']) "hack-o-rama"
		  ifTrue: [ stream next.
			    token _ stream peek ].
	      baseType _ self parseBaseType.
	      "### note that identifier may be missing"
	      varDecl _ self parseVarDecl.
	      decls addLast: 3. "!!! combine these into something and add it"
	      ].
    ^decls
! !

!CTypeParser methodsFor: 'private'!

init: aStream
    stream _ aStream.
! !

		      
    
    

