"======================================================================
|
|   A write stream which automatically indents
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:54:50 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
|
"

Stream subclass: #IndentedStream
	   instanceVariableNames: 'stream level amount indentString column'
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: nil
!


!IndentedStream class methodsFor: 'instance creation'!

on: aStream
    ^self new init: aStream

! !

!IndentedStream methodsFor: 'writing'!

nextPut: aChar
    column == 0 ifTrue: [ self indentLine ].
    stream nextPut: aChar.
    aChar == (Character nl) 
	ifTrue: [ column _ 0. ^self ].
    aChar == (Character tab) 
	ifTrue: [ column _ (column + 8) truncateTo: 8.  ^self ].
    column _ column + 1.
!

nextPutAll: aString
    | stringSize index start nl tab ch |
    index _ 1.
    stringSize _ aString size.
    nl _ Character nl.
    tab _ Character tab.
    [ index <= stringSize ] whileTrue:
	[ start _ index.
	  
	  [ (index <= stringSize) and: 
		[ ch _ aString at: index.
		  (ch ~= nl) and:
		      [ ch ~= tab ] ]]
	      whileTrue:
		  [ index _ index + 1 ].
	  column == 0 ifTrue: [ self indentLine ].
	  stream nextPutAll: (aString copyFrom: start to: (index - 1)).
	  column _ column + (index - start).
	  index <= stringSize  "must have found a newline"
	      ifTrue: 
		  [ self nextPut: ch.
		    index _ index + 1. ].
	  ].
! !



!IndentedStream methodsFor: 'accessing'!

level
    ^level
!

level: anInteger
    level _ anInteger.
    self computeIndent.
!

pushLevel
    self level: level + 1
!

popLevel
    self level: ((level - 1) max: 0)
!

indentAmount
    ^amount
!

indentAmount: anInteger
    amount _ anInteger.
    self computeIndent.
!

tabTo: col
    [ column < col ] whileTrue: 
	[ self space ].
! !



!IndentedStream methodsFor: 'standard operations'!

cr
    self nextPut: Character cr
!

nl
    self nextPut: Character nl.
!

crTab
    self cr.
    self tab
!

nlTab
    self nl.
    self tab
!

space
    self nextPut: Character space
!

tab
    self nextPut: Character tab
! !



!IndentedStream methodsFor: 'private'!

init: aStream
    stream _ aStream.
    amount _ 4.
    level _ 0.
    column _ 0.
    self computeIndent.
!

indentLine
    stream nextPutAll: indentString.
    column _ column + (amount * level).
!

computeIndent
    | space |
    space _ Character space.
    indentString _ String new: (amount * level).
    indentString atAllPut: space.
! !

