"======================================================================
|
|   Smalltalk in Smalltalk compiler
|
|   $Revision: 1.1 $
|   $Date: 1995/09/17 07:18:26 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	      3 Sep 95	  created.
|
"

FileStream fileIn: 'STSymTable.st'!


Compiler subclass: #STCompiler
       instanceVariableNames: 'methodNode symbolTable bytes abortBlock'
       classVariableNames: ''
       poolDictionaries: 'ByteCodeNames'
       category: nil
! 

!STCompiler class methodsFor: 'instance creation'!

onNodes: methodDefNode
    ^self new init: methodDefNode

! !


!STCompiler methodsFor: 'compilation'!

compile
    | selector body | 
    abortBlock _ [ ^self error: 'Aborted' ].
    selector _ node selector.
    selector args do:
	[ :anArg | symbolTable declareArgument: anArg ].
    
    body _ node body.
    body temporaries do:
	[ :aTemp | symbolTable declareTemporary: aTemp ].
    
    " ??? returned value?  more to do after compileBody?"
    self compileBody.
!


compileBody
    | body |
    body _ node body.
    body statements do: 
	[ :statement | self compileStatement: statement ].
    "??? return value?"
!

compileStatement: statementNode
    | byteCodes |
    (statementNode isMemberOf: ReturnNode)
	ifTrue: [ byteCodes _ self compileExpression: statementNode expression.
		  byteCodes addFirst: returnMethodStackTop.
		  ^byteCodes ]
	ifFalse: [ ^self compileExpression: statementNode ]
!

compileExpression: anExpressionNode
    | expr byteCodes |
    anExpressionNode isNil
	ifTrue: [ ^self notYetImplemented ].
    " !!! deal with anExpressionNode == nil"
    expr _ anEpxression expression.
    byteCodes _ self compileSimpleExpression: expr.
    byteCodes _ self compileAssignments: anExpression assigns
		     byteCodes: byteCodes.
    ^byteCodes
!

"Just a note in passing:
It seems like there should be some object that gets passed around
implictly which has the byte codes as an instance variable, but 
which has narrower scope than an entire method, but which has access
to the same symbol definitions.  You'd create a new one of these
to compile the contents of a block, possibly one for each statement or 
expression, and possibly for the inlined control flow structure blocks."

compileAssignments: assigns byteCodes: byteCodes
    | id variable |
    assigns do:			"Iteration order doesn't matter here"
	[ :assign | id _ assign id.
		    variable _ symbolTable lookup: id.
		    byteCodes _ self compileAssignentInto: variable byteCodes: byteCodes.
		    ].
    ^byteCodes
!

compileAssignentInto: variable byteCodes: byteCodes
    "Examine the variable -- if it's undefined, print an error message, cause 
     it to be defined, and continue."
    | symbolTableIndex definition locationType |
    variable isUnknown
	ifTrue: 
	    [ "Need to declare this puppy as a local, report an error, and go 
	       on"
	      self compileError: 'Undefined variable ', 
		  variable name printString, 'referenced.'.
	      
	      symbolTable declareTemporary: variable name.
	      variable _ symbolTalbe lookup: variable name.
	      "Fall through to the local variable case".
	      ].
    
    definition _ variable definition.
    variable isTemporary
	ifTrue: 
	    [ byteCodes addLast: DupStackTop.
	      definition <= 7
		  ifTrue: [ byteCodes addLast: 
				(PopTemporaryVariable bitOr: definition).
			    ^byteCodes ].
	     ].
    
    variable isInstVar
	ifTrue: 
	    [ byteCodes addLast: DupStackTop.
	      definition <= 7
		  ifTrue: [ byteCodes addLast: 
				(PopReceiverVariable bitOr: definition).
			    ^byteCodes ].
	     ].
		 
    locationType _ self computeLocationType: variable.
    byteCodes addLast: StoreIndexed;
	addLast: (locationType bitOr: definition)
    
    ^byteCodes
!

compileSimpleExpression: expr
    "GAK!  This should be more object oriented and use polymorphic dispatch,
     but then the compilation policy is distributed throughout a bunch of 
     objects which aren't really part of the compiler. "

    expr isMemberOf: IdentifierNode 
	 ifTrue: [ ^self compileVariable: expr ].
    ((expr isMemberOf: LiteralNode) | (expr isMemberOf: SymbolNode)
	 | (expr isMemberOf: ArrayConstNode))
	"Note -- this should be used to our advantage...different compilation
	 for the various things here "

    expr isMemberOf: BlockNode 
	 ifTrue: [ ^self compileBlock: expr ]
    expr isMemberOf: UnaryExprNode
	 ifTrue: [ ^self compileUnaryExpr: expr ].
    expr isMemberOf: BinaryExprNode
	 ifTrue: [ ^self compileBinaryExpr: expr ].
    expr isMemberOf: KeywordExprNode
	 ifTrue: [ ^self compileKeywordExpr: expr ].
    expr isMemberOf: CascadeNode 
	 ifTrue: [ ^self compileCascadedMessage: expr ].
	
    
    ^self compileExpression: expr
!






! !

!STCompiler methodsFor: 'Utility'!

computeLocationType: variable
    variable isTemporary ifTrue: [ ^TemporaryLocation ].
    variable isInstVar ifTrue: [ ^ReceiverLocation ].
    
    ^LiteralVariableLocation
! !



!STCompiler methodsFor: 'private'!

init: methodDefNode
    methodNode _ methodDefNode.
    symbolTable _ STSymbolTable new.
!

abort
    abortBlock value
! !


