"File BrowserMain.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date            Change
| bdiller       11 Oct 93       Created.
"
GuiData subclass:  #BrowserMain
        instanceVariableNames: 'windows'
        classVariableNames: 'shell '
        poolDictionaries: ''
        category: 'Graphics Windows'
!

!BrowserMain class methodsFor: 'accessing'!

close
"This method is invoked before quitting the browser and before saving the 
Smalltalk image.  When the system is launched subsequently, it is important 
that the shell be nil until the browser is initialized.  Other methods use the
state of this variable (shell) to probe the browser's initialization status"
    shell := nil.
!

shell
"Return application widget pointer.  This method is used to determine whether 
the X and browser environment is initialized.  If 'shell' is non-nil, the 
environment is completely initialized"
    ^shell
!!

!BrowserMain methodsFor: 'initializing'!

initialize
"Initialize X environment.  Create a button which will be used to operate the 
browser.  When the button is pressed, a pull down menu will be activated.  
The user can select the desired menu option"
| aMenu  win |
    Smalltalk at: #HandleErrorsWithGui put: true.
		PText initBuffer.
    Smalltalk enableGC:  true.
    blox := Blox bloxInit.
    shell := nil.
    Primitive initViews.
    windows := Set new.
    shell := TopLevelShell new: 'Smalltalk'.
    shell data: self.
    win := shell blox.
    win x: 10.
    win y: 10.
    
    win height: 30.
    win width: 80.
    
    win createToplevelWindow: 'window'.
    
    shell menu: ((Menu new: shell label: 'smalltalk') selectors:
        #(('worksheet' openWorksheet)('browser' openBrowser) 
         () ('save image' saveImage)('garbage collect' garbageCollect) 
         ()('quit...' quit)) 
                                                      receiver: self
                                                      argument: shell).
    aMenu := shell menu blox.
    aMenu x: 10.
    aMenu y: 10.
    
    aMenu height: 30.
    aMenu width: 80.
    
    shell display.
    blox dispatchEvents.
!!

!BrowserMain methodsFor: 'blue button messages'!

garbageCollect
"Force a full garbage collection in order to dispose of all unreferenced 
instances"
    Smalltalk compact.
!

openWorksheet
"Open a class hierarchy browser window.  This object registers itself in the 
Smalltalk dictionary.  It is not necessary to anchor this object onto an 
instance variable to prevent it from being GC'd "
    | worksheet textView |
    worksheet := TopLevelShell new: 'Worksheet'.
    worksheet addChildView: ((textView := PText new: worksheet)
       menuInit: ((PopupMenu new: textView label: '') selectors: 
      #(('cut' gstCut) ('copy' gstCopy) ('paste' gstPaste) 
       () ('clear' gstClear) () ('line...' line)  ('find...' find) 
       () ('do it' eval) ('print it' evalAndPrintResult) ('inspect' evalAndInspectResult) 
       () ('close' close))
       receiver: textView
       argument: nil);
       textMsg: #text;
       yourself).
    worksheet blox x: 0.
    worksheet blox y: 75.

    worksheet blox height: 175.
    worksheet blox width: 300.
    worksheet blox map.
!


openBrowser
    ClassHierarchyBrowser new openOn: Object.
!

quit
"Quit Smalltalk browser"
    | continue |
    windows notNil 
        ifTrue: [windows do: [ :w | w canClose ifFalse: [^shell beep]]].
    continue := false.
    ModalDialog new message: 'Save image before quitting?' in: shell;
        addButton: 'Yes' message: [self saveImage. continue := true];
        addButton: 'No' message: [continue := true];
        addButton: 'Cancel' message: [];
        display: shell.
    continue ifFalse: [^shell].
    blox release.
    blox terminateMainLoop.
!

saveImage
    | savedShell |
    "The class variable, 'shell', is used, secondarily as a flag to
     indicate initialization status.  Consequently, we need to nil out
     this variable before image is saved.  Otherwise, the browser may
     crash when attempting to display a Notifier or some other type of
     window before the X and Smalltalk system has been initialized"
    savedShell := shell.
    shell := nil.
    Smalltalk snapshot.
    shell := savedShell.
!!

!BrowserMain methodsFor: 'window maintenance'!

addWindow: toplevel
    windows add: toplevel.
!

removeWindow: toplevel
    windows remove: toplevel
!!

