"File DictInspect.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date                    Change
| bdiller               27 Oct 93       Created.
"

Inspector subclass:  #DictionaryInspector
        instanceVariableNames: ''
        classVariableNames: ''
        poolDictionaries: ''
        category: 'Graphics Windows'
!

!DictionaryInspector methodsFor: 'initializing'!

inspectMenu: listView
"Initialize menu for variable list pane"
    ^((PopupMenu new: listView label: '') selectors:
#(('inspect' evalAndInspectResult: listView) ('references' references: listView) 
() ('add field...' addField: listView) ('remove...' removeField: listView))
                                          receiver: self
                                          argument: listView ).
!!

!DictionaryInspector methodsFor: 'private'!

currentVariableValue
    currentVariable == 0
        ifTrue: [ ^nil ].
    ^(theObject at: currentVariable)
!

currentVariable: obj
    theObject at: currentVariable put: obj.
!

setInstanceVars: anObject
"Initialize instance variable, instVars, which governs display of
variable list pane.  Since the dictionary keys can be non-symbols as
well as symbols, it is not the case that the string representation of
the key will uniquely identify the dictionary entry.  Consequently,
the Association (dictionary entry) consisting of the key and value is
stored in the SortedCollection variable, instVars.  Using this
approach, we can pick out the Dictionary key's unique value by using
the selected list element's number as a key into instVars which is a
SortedCollection of Associations."
    theObject := anObject.
    base := 0.
    instVars := SortedCollection sortBlock: [ :a :b | (a isNil or: [ b isNil ])
                                                          ifTrue: [ true ]
                                                          ifFalse: [ a key <= b key ] ].
    theObject keysDo:
        [ :aKey | instVars add:  (Association key: (aKey printString asSymbol) 
                                              value: aKey)].
!!

!DictionaryInspector methodsFor: 'accessing'!

variable: assoc
"Set list selection to value of index.  Force a text view update"
    assoc key == 0
        ifTrue: [ currentVariable := 0 ]
        ifFalse: [currentVariable := (instVars at: assoc key) value].
    self changeState: #text.
    Primitive updateViews.
!

variableList
"Return sorted list of keys from set of Associations stored in instVars"
    ^instVars inject: (SortedCollection new) into: [:col :each | col add: each 
                                                                     key; yourself ].
!!

!DictionaryInspector methodsFor: 'variable list menu'!

addField: listView
"Prompt user for the name of new dictionary key.  If name is valid, add it
to dictionary"
    | key |
    listView canChangeState ifFalse: [^self].
    key := (Prompter message: 'Enter a new field' in: listView) response asSymbol.
    key ~= '' 
        "If new key already exists, reject "
        ifTrue: [(theObject includesKey: key ) ifTrue: [^ModalDialog new alertMessage: 
                                                            'Illegal name: the key, ' , key, ', already exists.' 
                                                                         in: listView] ]
        ifFalse: [^self].
    "Update variable selection"     
    currentVariable := key.
    "Update dictionary"
    theObject at: currentVariable put: nil.
    "Update instance variable governing variable list pane display"
    instVars add: (Association key: currentVariable value: currentVariable).
    "Update text view"
    self changeState: #variableList;
        changeState: #text.
    Primitive updateViews.
!

references: listView
"Open a method set browser on all methods which reference selected key"
    | alert keyRefs theKey |
    currentVariable isNil ifTrue: [^listView beep].
    keyRefs := SortedCollection new.
    Object withAllSubclasses do: 
        [:subclass | (subclass whichSelectorsReferTo: 
                          (theObject associationAt: currentVariable)) do: 
                              [ :sel | keyRefs add: ( subclass getName , ' ', sel )]].
    keyRefs isEmpty ifTrue: [^alert := ModalDialog new alertMessage: 
                                 'No references to ', (currentVariable printString) in: listView]. 
    MethodSetBrowser new openOn: keyRefs title: 'References to ', 
        (currentVariable printString)  selection: (currentVariable printString).
!

removeField: listView
| cancel |
    "Remove selected key from dictionary"
    currentVariable isNil ifTrue: [^listView beep].
     ModalDialog new message: ('Are you sure you want to remove, ', 
          currentVariable, '?') in: listView;
     addButton: 'Yes' message: [cancel := false];
     addButton: 'No' message: [cancel := true];
     display: listView.
     cancel ifTrue: [^self].
    "Remove key from dictionary"
    theObject removeKey: currentVariable.
    "Remove the association composed of the key and the value from the data object"
    instVars remove: (Association key: currentVariable value: currentVariable).
    currentVariable := 0.
    "Force a text view update to reflect deleted key"
    self changeState: #variableList;
        changeState: #text.
    Primitive updateViews.
!!
