"File ErrInspect.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date            Change
| bdiller       7 July 95.  Derived from sbb's BBrowser.st
|
"

Smalltalk at: #HandleErrorsWithGui put: true!

GuiData subclass: #ErrorInspector
       instanceVariableNames: 'callstackList stacktrace currentSelection numContexts errMessage'
       classVariableNames: ''
       poolDictionaries: ''
       category: nil
! 

!ErrorInspector class methodsFor: 'instance creation'!

new: message names: contextStrings contexts: contexts
    ^self new init: message names: contextStrings contexts: contexts
! !

!ErrorInspector methodsFor: 'accessing'!

currentContext
    currentSelection isNil ifTrue: [ ^nil ].
    ^callstackList at: currentSelection
!

next
    currentSelection isNil ifTrue: [ ^nil ].
    currentSelection _ currentSelection + 1 min: numContexts.
    "Should change the currently selected list item"
!

previous
    currentSelection isNil ifTrue: [ ^nil ].
    currentSelection _ currentSelection - 1 max: 1.
    "Should change the currently selected list item"
! !


!ErrorInspector methodsFor: 'callback'!

inspectContext
   self method
!

contextSelectedFrom: receiver at: anIndex
    currentSelection _ anIndex
! !

!ErrorInspector methodsFor: 'private'!
        
init: message names: nameList contexts: contexts
    | topLevel topView listView|
    stacktrace _ nameList.
    numContexts _ contexts size.
    errMessage _ message.
    topView _ ((TopLevelShell new: message ) data: self).
    topLevel _ topView blox.
    
    topView addChildView: ((listView := PList new: 'MethodSet' in: topView)
			       initialize;
			       data: self;
			       listMsg: #stacktrace; 
			       handleUserChange: #contextSelectedFrom:; 
			       menuInit: ((PopupMenu new: listView label: '') 
					     selectors: #((proceed)
					     ('inspect' inspectContext)
					     ()
				             ('trace dump' traceDump))
					      receiver: self
	                		      argument: listView;
					      selectors: #(() ('close' close))
					      receiver: listView
					      argument: nil;
                                              yourself);
			       yourself).

    callstackList _ listView blox.
    callstackList width: 300; height: 100.
    
    callstackList contents: nameList elements: contexts.
    
    topView display.
!

close: aView
    | tv |
    tv := aView rootView blox.
    aView rootView close ifTrue: [tv destroy: (tv guiObject)].
!!



!ErrorInspector methodsFor: 'callback'!

traceDump
    | aFile |
    aFile := FileStream open: 'TraceDump' mode: 'a'.
    aFile nextPutAll: errMessage; nl.
    0 to: callstackList numberOfStrings - 1 do: [ :i | aFile nextPutAll: (callstackList labelAt: i) asString; nl].
    aFile nl.
    aFile close
!

stacktrace
	^stacktrace
!

inspectContext
    ContextInspector new: self.
!

contextSelectedFrom: assoc
    currentSelection _ assoc key - 1.
! !



(Object includesSelector: #originalDoesNotUnderstand: )
    ifFalse: [ Object addSelector: #originalDoesNotUnderstand: 
         withMethod: (Object compiledMethodAt: #doesNotUnderstand:) ]!


!Object methodsFor: 'overriding'!

error: message
    | context names contexts name nameStr |
    "Check for blox init being called already!"
    names _ OrderedCollection new.
    contexts _ OrderedCollection new.
    
    context _ thisContext.
        
    [ context notNil ] whileTrue: 
        [ nameStr _ WriteStream on: (String new: 5).
          nameStr nextPutAll: context realReceiver;
              nextPutAll: '>>#'; 
              nextPutAll: context selector printString.
          name _ nameStr contents.
          names add: name.
          contexts add: context.
          context _ context parentContext.
          ].
"self print. ' ' asSymbol print. message asSymbol printNl.
                Smalltalk backtrace."
    ErrorInspector new: message names: names contexts: contexts.
!

doesNotUnderstand: message
    (HandleErrorsWithGui )
        ifTrue: [ self error: 'Did not understand ', message selector. ]
        ifFalse: [ self originalDoesNotUnderstand: message ].
! !


!BlockContext methodsFor: 'debugging'!

callers
    self inspect.
    caller notNil
        ifTrue: [ caller callers ]
!

parentContext
    ^caller
!

selector
    ^selector
!

receiverName
    ^'[] in ', home class name
!

realReceiver
    "Temp hack"
    ^'[] in ', home class name

! !

!MethodContext methodsFor: 'debugging'!

callers
    self inspect.
    sender notNil 
        ifTrue: [ sender callers ]
!

parentContext
    ^sender
!

receiverName
    | name |
    name _ receiver class name.
    name isNil ifTrue: [ name _ receiver name ].
    ^name
!

realReceiver
    | name receiverClass methodClassName |
    
    receiverClass _ receiver class.
    receiverClass ~~ methodClass
        ifTrue: [ name _ self receiverName.
                  methodClassName _ methodClass name.
                  ^name , '(', methodClassName, ')'  ]
        ifFalse: [ ^self receiverName ].
! !


