"======================================================================
|
|   16 bit String Method Definitions
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
 ======================================================================"


"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	      7 Dec 91	  Created from String.st
|
"

ArrayedCollection variableSubclass: #String16
		  instanceVariableNames: ''
		  classVariableNames: ''
		  poolDictionaries: ''
		  category: nil
!

String16 comment: 
'My instances represent 16-bit string data types.  I provide accessing and
manipulation methods for 16-bit string data types.' !

!String16 class methodsFor: 'basic'!

fromString: aString
    "Creates a 16 bit string from an 8 bit one.  I'm not sure that this
     is really right, but it works for what I need currently.  The 
     8 bit characters are interpreted to be the low character of char 2 b's"
    | str |
    str _ self new: aString size.
    1 to: aString size do:
	[ :i | str at: i
		   put: (Char2B byte1: 0
				byte2: (aString at: i) asciiValue) ]. 
    ^str
! !



!String16 methodsFor: 'comparing'!

< aString16
    "Return true if the receiver is less than aString16, ignoring case
    differences."
    self >= aString16 ifTrue: [ ^false ]
    	    	    ifFalse: [ ^true ]
!

> aString16
    "Return true if the receiver is greater than aString16, ignoring case
    differences."
    self <= aString16 ifTrue: [ ^false ]
    	    	    ifFalse: [ ^true ]
!

<= aString16
    "Returns true if the receiver is less than or equal to aString16,
    ignoring case differences.  If is receiver is an initial substring of
    aString16, it is considered to be less than aString16."
    | c1 c2 |
    " Scan self and aString16 until a character is clearly greater or lesser
      (All preceding characters must have been equal).  If the end is reached,
      one of the strings is a possibly improper initial substring of the other,
      and for the receiver to be less than aString16, it must be the initial
      substring."
    1 to: (self size min: aString16 size) do:
    	[ :i | c1 _ (self at: i) asLowercase.
	       c2 _ (aString16 at: i) asLowercase.
	       c1 < c2 ifTrue: [ ^true ].
	       c1 > c2 ifTrue: [ ^false ] ].
    ^self size <= aString16 size
!

>= aString16
    "Returns true if the receiver is greater than or equal to aString16,
    ignoring case differences.  If is aString16 is an initial substring of
    the receiver, it is considered to be less than the receiver."
    | c1 c2 |
    1 to: (self size min: aString16 size) do:
    	[ :i | c1 _ (self at: i) asLowercase.
	       c2 _ (aString16 at: i) asLowercase.
	       c1 < c2 ifTrue: [ ^false ].
	       c1 > c2 ifTrue: [ ^true ] ].
    ^self size >= aString16 size
!



sameAs: aString16
    "Returns true if the receiver is the same string as aString16, nominally 
     ignoring case differences.  In reality, I don't know much about 16 bit 
     strings, so I don't know which are lowercase.  But, I'll leave this
     decision to the Char16 datatype (heh heh heh)."
    self size ~= aString16 size ifTrue: [ ^false ].
    1 to: self size do:
    	[ :i | (self at: i) asLowercase ~= (aString16 at: i) asLowercase
	    	ifTrue: [ ^false ] ].
    ^true
! !



!String16 methodsFor: 'converting'!

asUppercase
    "Returns a copy of self as an uppercase string"
    | newStr |
    newStr _ self species new: self size.
    1 to: self size do:
    	[ :i | newStr at: i put: (self at: i) asUppercase ].
    ^newStr
!

asLowercase
    "Returns a copy of self as a lowercase string"
    | newStr |
    newStr _ self species new: self size.
    1 to: self size do:
    	[ :i | newStr at: i put: (self at: i) asLowercase ].
    ^newStr
! !



!String16 methodsFor: 'copying'!

shallowCopy
    | newStr |
    newStr _ self species new: self size.
    1 to: self size do:
    	[ :i | newStr at: i put: (self at: i) ].
    ^newStr
!

deepCopy
    | newStr |
    newStr _ self species new: self size.
    1 to: self size do:
    	[ :i | newStr at: i put: (self at: i) deepCopy ].
    ^newStr
!

, aString16
    | newString mySize |
    newString _ self species new: ((mySize _ self size) + aString16 size).
    newString replaceFrom: 1 to: self size with: self.
    newString replaceFrom: self size + 1 to: self size + aString16 size 
	      with: aString16.
    ^newString
!

replaceFrom: start to: stop with: replacementCollection
    self replaceFrom: start to: stop with: replacementCollection 
	 startingAt: 1
!

replaceFrom: start to: stop with: replacementCollection startingAt: repStart
    | len |
    len _ stop - start + 1.
    len ~= replacementCollection size
	ifTrue: [ ^self error: 'replacementCollection has different size' ].
    1 to: len do:
	[ :i | self at: start + i - 1 put: (replacementCollection at: i) ]
! !

