"======================================================================
|
|   X Server Authorization support
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
 ======================================================================"


"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     28 Aug 94	  Fixed a bug in bestMatchFor: ; it wasn't returning a
|			  nullRec in all cases.
|
| sbb	      4 Jul 92	  Now handles non-existent .Xauthority properly.
|
| sbb	     23 Feb 92	  created.
|
"


X subclass: #XAuth
  instanceVariableNames: ''
  classVariableNames: 'FamilyLocal FamilyWild'
  poolDictionaries: ''
  category: 'Experimental'
!

X subclass: #XAuthRec
  instanceVariableNames: 'family addrLen addr dpyLen dpyStr nameLen
                          name dataLen data'
  classVariableNames: ''
  poolDictionaries: ''
  category: 'Experimental'
!

UnixStream subclass: #XAuthStream
	   instanceVariableNames: ''
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: 'Experimental'
!


XAuth comment: 
'I am an early attempt to provide access to the X authority mechanism.
I''m not as robust as I should be, but I will probably work in most cases.
If you feel like extending my methods, I would surely appreciate it. 

Thanks.
'!

!XAuth class methodsFor: 'connection authorization'!

authFilename
    | name home |
    name _ Smalltalk getenv: 'XAUTHORITY'.
    name notNil ifTrue: [ ^name ].

    home _ Smalltalk getenv: 'HOME'.
    home isNil ifTrue: [ home _ '' ].
    ^home , '/.Xauthority'
!

bestMatchFor: hostName
    "Returns the authorization object for the given host.  If there
     is none, it returns a null authorization object"
    "Currently, UNIX/TCP specific, but please feel free to extend"
    | filename stream rec |
    filename _ XAuth authFilename.
    stream _ XAuthStream onFile: filename.
    stream notNil 
	ifTrue:	[
	    [ rec _ stream next.
	      rec notNil ] whileTrue:
		  [ ((hostName = '') and: [ rec family == FamilyLocal ])
			ifTrue: [ stream close. 
				  ^rec ].
		    ((hostName ~= '') and: [ hostName asByteArray = rec addr ])
			ifTrue: [ stream close.
				  ^rec ].
		    " Should add some test based on getting the host address "
		    rec ].
	    stream close. ].
    
    ^XAuthRec nullRec
!

test
    | filename stream rec |
    filename _ XAuth authFilename.
    stream _ XAuthStream onFile: filename.
    [ rec _ stream next.
      rec notNil ] whileTrue:
	  [ rec inspect ].
    stream close
!

initialize
    FamilyLocal _ 256.
    FamilyWild _ 65535.
! !

!XAuthStream class methodsFor: 'instance creation'!

onFile: aFileName
    ^self open: aFileName dir: UnixStream readOnly
! !



!XAuthStream methodsFor: 'accessing'!

next
    ^XAuthRec readFrom: self
! !



!XAuthRec class methodsFor: 'instance creation'!

readFrom: aStream
    | obj |
    obj _ self new.
    (obj initFrom: aStream)
	isNil ifTrue: [ ^nil ].
    ^obj
!

nullRec
    ^self new initNull
! !



!XAuthRec methodsFor: 'accessing'!

family
    ^family
!

addr
    ^addr
!

display
    ^dpyStr
!

name
    ^name
!

data
    ^data
! !



!XAuthRec methodsFor: 'private'!

initFrom: aStream
    (family _ aStream readShort) isNil
	ifTrue: [ ^nil ].
    addrLen _ aStream readShort.
    addr _ ByteArray new: addrLen.
    aStream read: addr.
    dpyLen _ aStream readShort.
    dpyStr _ ByteArray new: dpyLen.
    aStream read: dpyStr.
    nameLen _ aStream readShort.
    name _ ByteArray new: nameLen.
    aStream read: name.
    dataLen _ aStream readShort.
    data _ ByteArray new: dataLen.
    aStream read: data
!

initNull
    addrLen _ 0.
    addr _ ByteArray new: 0.
    dpyLen _ 0.
    dpyStr _ ByteArray new: 0.
    nameLen _ 0.
    name _ ByteArray new: 0.
    dataLen _ 0.
    data _ ByteArray new: 0
! !

XAuth initialize!


"Some self testing code"
"
(XAuth bestMatchFor: '') inspect.
(XAuth bestMatchFor: 'rambo') inspect.
(XAuth bestMatchFor: 'laplace') inspect.
!
"

