"======================================================================
|
|   STIX basic X Event mechanism
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

"======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"


"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbyrne     24 May 90	  created.
|
"

XProtocol subclass: #XEvent
	  instanceVariableNames: 'sequenceNumber'
	  classVariableNames: 'EventMap'
	  poolDictionaries: 'XGlobals'
	  category: 'X hacking'
!

FileStream fileIn: 'Xevt.st'!

"Fix up those events which can't be generated automatically."

!ClientMessageEvent methodsFor: 'Serialization'!

eventTo: aStream
    aStream byte: 33;
	ubyte: format; 
	uword: sequenceNumber; 
	ulong: window; 
	ulong: type.
    format = 8 ifTrue: 
	[ data size ~= 20
	      ifTrue: [ ^self error: 'ClientMessage not 20 bytes long' ].
	  aStream bytes: data ].
    format = 16 ifTrue: 
	[ data size ~= 10
	      ifTrue: [ ^self error: 'ClientMessage not 10 words long' ].
	  data do: [ :item | aStream uword: item ] ].
    format = 32 ifTrue: 
	[ data size ~= 5
	      ifTrue: [ ^self error: 'ClientMessage not 5 longs long' ].
	  data do: [ :item | aStream ulong: item ] ].
! !



!ClientMessageEvent methodsFor: 'private'!

initFrom: aStream
    format _ aStream ubyte. 
    sequenceNumber _ aStream uword. 
    window _ aStream ulong. 
    type _ aStream ulong. 
    format == 8
	ifTrue: [ data _ aStream getBytes: 20 ].
    format == 16
	ifTrue: [ data _ Array new: 10.
		  1 to: 10 do: 
		      [ :i | data at: i 
				  put: (aStream uword) ] ].
    format == 32
	ifTrue: [ data _ Array new: 5.
		  1 to: 5 do: 
		      [ :i | data at: i 
				  put: (aStream ulong) ] ].
! !

"--- End of custom event code ---"




!XEvent class methodsFor: 'subinstance creation'!

initialize
    | namePrefixes i sym |
    namePrefixes _ #(
KeyPress			"2"
KeyRelease			"3"
ButtonPress			"4"
ButtonRelease			"5"
MotionNotify			"6"
EnterNotify			"7"
LeaveNotify			"8"
FocusIn				"9"
FocusOut			"10"
KeymapNotify			"11"
Expose				"12"
GraphicsExposure		"13"
NoExposure			"14"
VisibilityNotify		"15"
CreateNotify			"16"
DestroyNotify			"17"
UnmapNotify			"18"
MapNotify			"19"
MapRequest			"20"
ReparentNotify			"21"
ConfigureNotify			"22"
ConfigureRequest		"23"
GravityNotify			"24"
ResizeRequest			"25"
CirculateNotify			"26"
CirculateRequest		"27"
PropertyNotify			"28"
SelectionClear			"29"
SelectionRequest		"30"
SelectionNotify			"31"
ColormapNotify			"32"
ClientMessage			"33"
MappingNotify			"34"
).
    EventMap _ Array new: namePrefixes size + 1.
    i _ 2.
    namePrefixes do: 
	[ :prefix | sym _ (prefix, 'Event') asSymbol.
		    EventMap at: i put: (Smalltalk at: sym 
						   ifAbsent: [ nil ]).
		    i _ i + 1 ]
!

newFrom: aStream type: type
    | class | 
    "---> Fix to handle if the subscript is out of bounds.
     
     Type will have the high order bit (128) turned on if the event was from 
     send event and not an event generated by the server"
    type _ type bitAnd: 127.
    
    (type between: 1 and: EventMap size)
	 ifFalse: [ ^self error: 'Event Type ', type printString, 
			' out of range' ].
    class _ EventMap at: type.
    ^class newEvtFrom: aStream
!

newEvtFrom: aStream
    ^self new initFrom: aStream
!!



!XEvent methodsFor: 'accessing'!

sequenceNumber
    ^sequenceNumber
!

mappedId: anId
    ^anId
"Registry isn't implemented
    ^Registry at: anId
	      ifAbsent: [ nil ]
"
!!

XEvent initialize!


LoadStatus _ 'XEvent'!
