"======================================================================
|
|   Superclass for all of the variable arity X request operations.
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"

 "======================================================================
|
| Copyright (C) 1990, 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	      4 Apr 93	  Added shared instance and byte vector to reduce
|			  consing and improve performance.
|
| sbb	      9 Nov 90	  Created.
|
"

XProtocol subclass: #XRequest
	  instanceVariableNames: 'stream pos values'
	  classVariableNames: 'sharedInstance commonInstance'
	  poolDictionaries: ''
	  category: 'X request container'
!

XRequest comment:
'I provide methods for creating and issuing requests to the X server.  Some
of my methods encode all of the generic datatypes for sending to the server,
and I provide auxiliary functions for adhereing to the X protocol specification
(such as doing padding to a 4 byte boundary).' !

(Smalltalk includesKey: #Font)
    ifFalse: [ Smalltalk at: #Font put: nil ]
!



!XRequest class methodsFor: 'instance creation'!

command: aByte
    ^self command: aByte aux: 0
!

command: aByte aux: aByte2
    ^commonInstance init: aByte aux: aByte2
"    ^(self new) init: aByte aux: aByte2"
!

fixedCommand: aByte
    ^self fixedCommand: aByte aux: 0
!

fixedCommand: aByte aux: aByte2
    ^sharedInstance initFixed: aByte aux: aByte2
!

simpleRequest
    ^self new simpleInit
! !



!XRequest class methodsFor: 'initialization'!

initialize
    sharedInstance _ self new.
    sharedInstance initStream.
    commonInstance _ self new
! !



!XRequest methodsFor: 'X request users'!

byte: aByte
    stream nextPut: aByte
!

ubyte: aByte
    aByte class == Character
	ifTrue: [ stream nextPut: aByte asciiValue ]
	ifFalse: [ stream nextPut: aByte ]	"hack for now"
!

word: aWord
    Bigendian 
	ifTrue: [ stream nextPut: ((aWord bitShift: -8) bitAnd: 255).
		  stream nextPut: (aWord bitAnd: 255) ]
	ifFalse: [ stream nextPut: (aWord bitAnd: 255).
		   stream nextPut: ((aWord bitShift: -8) bitAnd: 255). ]
!

uword: aWord
    Bigendian 
	ifTrue: [ stream nextPut: aWord // 256.
		  stream nextPut: (aWord bitAnd: 255) ]
	ifFalse: [ stream nextPut: (aWord bitAnd: 255).
		   stream nextPut: aWord // 256. ]
!

long: aLong
    Bigendian 
	ifTrue: [ self uword: ((aLong bitShift: -16) bitAnd: 65535).
		  self uword: (aLong bitAnd: 65535) ]
	ifFalse: [ self uword: (aLong bitAnd: 65535).
		   self uword: ((aLong bitShift: -16) bitAnd: 65535). ]

!

ulong: aLong
    Bigendian 
	ifTrue: [ self uword: aLong // 65536. 
		  self uword: (aLong bitAnd: 65535) ]
	ifFalse: [ self uword: (aLong bitAnd: 65535).
		   self uword: aLong // 65536 ]

!

bool: aBool
    aBool			"### May extend this to support #True & #False"
	ifTrue: [ stream nextPut: 1 ]
	ifFalse:[ stream nextPut: 0 ]
!

bytes: byteArray
    stream nextPutAll: byteArray
!

string: aString
    aString do: [ :ch | stream nextPut: ch asciiValue ]
!

str: aString
    stream nextPut: aString size.
    self string: aString.
!

char2B: aChar2B
    self byte: aChar2B byte1;
	byte: aChar2B byte2
!

string16: aString16
    aString16 do: [ :ch2b | self char2B: ch2b ].
!

keysym: aKeysym
    self ulong: aKeysym
!

colorItem: aColorItem
    | bits |
    self ulong: aColorItem pixel;
	uword: aColorItem red;
	uword: aColorItem green;
	uword: aColorItem blue.
    bits _ 0.
    aColorItem doRed   ifTrue: [ bits _ bits bitOr: 1].
    aColorItem doGreen ifTrue: [ bits _ bits bitOr: 2].
    aColorItem doBlue  ifTrue: [ bits _ bits bitOr: 4].
    self byte: bits.
    self byte: 0.		"unused"
!

zeroBytes: numBytes
    "crude, but effective"
    numBytes timesRepeat: [ self byte: 0 ]
!

queueOn: server
    | length |
    self pad.
    length _ stream position // 4.
    stream position: 3.
    self word: length.
    ^server queueRequest: stream contents
!

simpleContents
    ^stream contents
!!




!XRequest methodsFor: 'subclass methods'!

at: index byte: aValue
    ^self at: index long: aValue
!

at: index bool: aValue
    ^self at: index 
	  long: (aValue ifTrue: [ 1 ] ifFalse: [ 0 ])
!

at: index ubyte: aValue
    ^self at: index long: aValue
!

at: index word: aValue
    ^self at: index long: aValue
!

at: index uword: aValue
    ^self at: index long: aValue
!

at: index ulong: aValue
    ^self at: index long: aValue
!

at: index long: aValue
    values at: index put: aValue
!


noBits
    values _ Array new: 32
!

emitBitValues: isWord
    | bitMask |
    bitMask _ 0.
    values inject: 1 into: 
	[ :bit :value | value notNil
			    ifTrue: [ bitMask _ bitMask bitOr: bit ].
			bit * 2 ].
    isWord
	ifTrue: [ self uword: bitMask; uword: 0 ]
	ifFalse: [ self ulong: bitMask ].
    values do: [ :value | value notNil
			      ifTrue: [ self long: value ] ]
! !



!XRequest methodsFor: 'higher level object encoding'!

point: aPoint
    self word: aPoint x.
    self word: aPoint y
!

rectangle: aRectangle
    self point: aRectangle origin.
    self point: aRectangle extent
!

arc: anArc
    self word: anArc x.
    self word: anArc y.
    self uword: anArc width.
    self uword: anArc height.
    self word: anArc angle1.
    self word: anArc angle2.
!

segment: aSegment
    self word: aSegment x1.
    self word: aSegment y1.
    self word: aSegment x2.
    self word: aSegment y2
!

textItem8: aTextElt
    | id |
    aTextElt class == Font
	ifTrue: [ self ubyte: 255. "font shift indicator"
		  id _ aTextElt id.
		  self ubyte: ((id bitShift: -24) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -16) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -8) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -0) bitAnd: 16rFF) ]
	ifFalse: [ self ubyte: aTextElt string size.
		   self byte: aTextElt delta.
		   self string: aTextElt string ]
!

textItem16: aTextElt
    | id |
    aTextElt class == Font
	ifTrue: [ self ubyte: 255. "font shift indicator"
		  id _ aTextElt id.
		  self ubyte: ((id bitShift: -24) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -16) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -8) bitAnd: 16rFF).
		  self ubyte: ((id bitShift: -0) bitAnd: 16rFF) ]
	ifFalse: [ self ubyte: aTextElt string size.
		   self byte: aTextElt delta.
		   self string16: aTextElt string ]
!

pad
    | curPos numBytes |
    curPos _ stream position.
    numBytes _ 4 - (curPos - 1) bitAnd: 3.

    stream next: numBytes put: 0 
!!




!XRequest methodsFor: 'private'!

initFixed: aByte aux: aByte2
    stream emptyStream.
    self byte: aByte.
    self byte: aByte2.
    self word: 0.		"placeholder for length"
    self localInit
!

init: aByte aux: aByte2
    "stream _ ReadWriteStream onEmpty: (ByteArray new: 32)."
    stream _ FastRWStream onEmpty: (ByteArray new: 32).
    self byte: aByte.
    self byte: aByte2.
    self word: 0.		"placeholder for length"
    self localInit
!

simpleInit
    "stream _ ReadWriteStream onEmpty: (ByteArray new: 32)"
    stream _ FastRWStream onEmpty: (ByteArray new: 32)
!

localInit
    ^self
!

initStream
    "stream _ ReadWriteStream onEmpty: (ByteArray new: 32)"
    stream _ FastRWStream onEmpty: (ByteArray new: 32)
! !

XRequest initialize!

LoadStatus _ 'XRequest'!

FileStream fileIn: 'XKeyboardRequest.st'!
FileStream fileIn: 'XGCRequest.st'!
FileStream fileIn: 'XWindowRequest.st'!
FileStream fileIn: 'XConfigureWindowRequest.st'!
