"======================================================================
|
|   A very simple game for a two year old.
|   'Here, Bobby, this called a MOUSE...'
|   Evaluate the commented code at the bottom to check this out.
|
|   $Revision: 1.3 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     28 Aug 94	  Added header stuff.
|
"



Object subclass: #Game
       instanceVariableNames: 'redPix bluePix greenPix objects'
       classVariableNames: 'window'
       poolDictionaries: ''
       category: nil
!

FileStream fileIn: 'WinEvents.st'!
FileStream fileIn: 'Shape.st'!



!Game class methodsFor: 'instance creation'!

new
    ^super new init
! !


!Game methodsFor: 'operation'!

run
    window server processEvents.
!

stopGame
    self playSound: 'bong.au'.
    window server close.
    ExitHandlerLoop _ true.
! !


!Game methodsFor: 'assistance'!

drawGame
    objects do: 
	[ :obj | obj drawOn: window ].
!

playSound: soundName
    (Smalltalk hasFeatures: #sparc)
	ifTrue: [ Smalltalk system: 'cp ', soundName ,' /dev/audio'. ] 
!

doinkBackground
    | blackPixel whitePixel |
    self playSound: 'drip.au'.

 "
    whitePixel _ (window server root screens at: 1) whitePixel.
    blackPixel _ (window server root screens at: 1) blackPixel.

    5 timesRepeat: [ window changeWindowAttributes: 
			 [ :r | r backgroundPixel: blackPixel ].
		     window changeWindowAttributes: 
			 [ :r | r backgroundPixel: whitePixel ] ].
 "
! !


!Game methodsFor: 'event handling'!

handleButtonPress: event
    | point |
    point _ event eventX @ event eventY.
    objects do: 
	[ :obj | (obj containsPoint: point) 
		     ifTrue: [ obj eraseFrom: window.
			       self playSound: (#('cuckoo.au' 
						      'laugh.au' 
						      'train.au') 
						    at: event detail).
			       objects remove: obj.
			       objects size == 0
				   ifTrue: [ self stopGame ].
			       ^self ] ].
    self doinkBackground.
!

handleKeyPress: event
    event inspect.
    (event state == 5 "c-s-")
	ifTrue: [   "control = 83, shift = 106"
		     self stopGame ].
"	ifFalse: [ self playSound: 'bark.au' ]"
! !



!Game methodsFor: 'private'!

init
    | s root screen rootWin cmap cursor |
    window notNil
	ifTrue: [ window server close.
		  " ### Very clumsy "
		  EventHandlers _ Dictionary new. ].
    s _ Server connectToDisplay.
    root _ s root.
    screen _ root screens at: 1.
    rootWin _ screen window.

    cursor _ self newCursor: s.

    window _ Window createWindow: rootWin class: #CopyFromParent
		    depth: 0
		    visual: #CopyFromParent x: 0 y: 0
		    width: screen widthInPixels height: screen heightInPixels
		    borderWidth: 5 onServer: s
		    attrs: [ :r | r eventMask: #(Exposure ButtonPress KeyPress);
				      backgroundPixel: screen whitePixel;
				      cursor: cursor ].
    cmap _ screen defaultColormap.
    window name: 'Bobby''s Game!'.

    redPix _ (cmap allocNamedColor: 'red') pixel.
    bluePix _ (cmap allocNamedColor: 'blue') pixel.
    greenPix _ (cmap allocNamedColor: 'green') pixel.
    self registerEventHandlers.    
    self createObjects.
    window mapWindow.

!

newCursor: server
    | f cur |
    f _ Font openFont: 'cursor' onServer: server.
    cur _ Cursor createGlyphCursor: f maskFont: #None
		 sourceChar: 21 maskChar: 0
		 foreRed: 0 foreGreen: 0 foreBlue: 0
		 backRed: 16rFFFF backGreen: 16rFFFF backBlue: 16rFFFF
		 onServer: server.
    f closeFont.
    ^cur
!

registerEventHandlers
    window registerFor: ExposeEvent
	   when: [ :e | e window == window id ]
	   do: [ :e | e count = 0
			  ifTrue: [ self drawGame ] ].
    window registerFor: ButtonPressEvent
	   when: [ :e | e event == window id ]
	   do: [ :e | self handleButtonPress: e ].
    
    window registerFor: KeyPressEvent
	   when: [ :e | e event == window id ]
	   do: [ :e | self handleKeyPress: e ].
    
    " ### more "
!

createObjects
    | gc rect size gap | 
    objects _ OrderedCollection new: 0.

    size _ 300.
    gap _ 50.
    rect _ 0@0 corner: size @ size.
    

    rect moveBy: 50@200.
    gc _ GC createGC: window onServer: window server 
	    args: [ :r | r foreground: redPix ].
    objects add: (Shape new: gc bounds: rect copy).
    
    rect moveBy: (size + gap) @ 0.
    gc _ GC createGC: window onServer: window server 
	    args: [ :r | r foreground: bluePix ].
    objects add: (Triangle new: gc bounds: rect copy).

    rect moveBy: (size + gap) @ 0.
    gc _ GC createGC: window onServer: window server 
	    args: [ :r | r foreground: greenPix ].
    objects add: (Circle new: gc bounds: rect copy).

! !

 
"Test area"

"
| g |
    g _ Game new.
    g run
!
"
