"======================================================================
|
|   Lego editing code.  Cute, shows promise, not the best implementation.
|   Left mouse places, middle mouse moves last placed piece, right mouse
|   exits.
|
|   $Revision: 1.4 $
|   $Date: 1995/06/06 14:56:15 $
|   $Author: sbb $
|
 ======================================================================"


"======================================================================
|
| Copyright (C) 1991, 1992, 1994, 1995 Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================"

"
|     Change Log
| ============================================================================
| Author       Date       Change 
| sbb	     20 Sep 94	  Incorporated Joerge Rade's fix to button 2 handling.
|
| sbb	     30 Aug 94	  added header stuff.
|
"
 

FileStream fileIn: 'Lego.st'!
Smalltalk at: #EventHandlers put: Dictionary new!
Smalltalk at: #ExitHandlerLoop put: nil!
Smalltalk at: #AccessorMethods put: Dictionary new!

!Window methodsFor: 'event handling'!

registerFor: anEventClass when: conditionBlock do: handlerBlock
    | handlers |
    handlers _ EventHandlers at: anEventClass
				ifAbsent: [ EventHandlers at: anEventClass
							  put: OrderedCollection new.
					    EventHandlers at: anEventClass ].
    handlers add: conditionBlock -> handlerBlock
! !


!Server methodsFor: 'event dispatching'!

processEvents
    | e win handlers |
    ExitHandlerLoop _ false.
    [ ExitHandlerLoop ] whileFalse:
	[ e _ self nextEvent.
	  handlers _ EventHandlers at: e class
				   ifAbsent: 
				       [ ^self error: 'Could not find a handler for ',
					     e class printString ].
	  handlers do:
	      [ :handler | (handler key value: e)
			       ifTrue: [ handler value value: e ]
			       ]
	      ]
! !

Lego subclass: #DemoLego
     instanceVariableNames: 'width size'
     classVariableNames: ''
     poolDictionaries: ''
     category: 'Fun hacking'
!

!DemoLego class methodsFor: 'fun hacking'!

createMainWindowOn: s
    ^self new init: s
! !

!DemoLego methodsFor: 'fun hacking'!

width: aWidth
    width _ aWidth
!	

size: aSize
    size _ aSize
! !

!DemoLego methodsFor: 'private'!

init: s
    | w root screen rootWin |
    root _ s root.
    screen _ root screens at: 1.
    rootWin _ screen window.
    width _ 1.
    size _ 6.
    w _ Window createWindow: rootWin class: #CopyFromParent 
	       depth: 0
	       visual: #CopyFromParent x: 0 y: 0
	       width: 400 height: 300 borderWidth: 5 onServer: s
	       attrs: [ :r | r eventMask: #(Exposure ButtonPress);
				 backgroundPixel: screen blackPixel ].
    w changeProperty: (Atom prop: #WmName on: s)
      type: (Atom prop: #String on: s)
      format: 8
      mode: #Replace
      data: 'Legos'.
    w inspect.
    w mapWindow.
    s nextEvent inspect.		"ignore expose event"
    self setupHandlersFor: w
!

setupHandlersFor: w
    | last point bytes |
    last _ OrderedCollection new.
    w registerFor: ButtonPressEvent
      when: [ :e | e event == w id ]
      do: [ :e | "'got an event' printNl."
		 point _ e eventX @ e eventY. 
		 point _ point grid: ( 15 @ (3 * size) ).
		 e detail == 3
		     ifTrue: [ ExitHandlerLoop _ true ].
		 e detail == 1
		     ifTrue: [ bytes _ Smalltalk byteCodeCounter.
			       last addLast: (Lego legoAt: point
						   width: width height: 1 
						   color: 'white' on: w
						   vScale: size).
			       "stdout nextPutAll: 'Bytes executed: ';
				      print: Smalltalk byteCodeCounter - bytes;
				      nl."
				   ].
		 ((e detail == 2) and: [last size > 0])
			  ifTrue: 
			      [ last last moveTo: point.
				last do: [ :lego | lego draw ]
				    ]
		     ].
    w registerFor: ExposeEvent
      when: [ :e | e window == w id ]
      do: [ :dummy | last do: [ :lego | lego draw ] ].


! !

Lego subclass: #Pieces
     instanceVariableNames: 'w s contents prevPiece drawingInst'
     classVariableNames: 'Gap YStart'
     poolDictionaries: ''
     category: 'Fun hacking'
!

!Pieces class methodsFor: 'instance creation'!

createWindowOn: server onInst: inst
    ^self new init: server forInst: inst
!

initialize
    Gap _ 30.
    YStart _ 20.
! !

Pieces initialize!

!Pieces methodsFor: 'private'!

init: server forInst: inst
    | root screen rootWin |
    s _ server.
    drawingInst _ inst.
    root _ s root.
    screen _ root screens at: 1.
    rootWin _ screen window.
    w _ Window createWindow: rootWin class: #CopyFromParent 
	       depth: 0
	       visual: #CopyFromParent x: 0 y: 0
	       width: 150 height: 300 borderWidth: 5 onServer: s
	       attrs: [ :r | r eventMask: #(Exposure ButtonPress);
				 backgroundPixel: screen blackPixel ].
    w changeProperty: (Atom prop: #WmName on: s)
      type: (Atom prop: #String on: s)
      format: 8
      mode: #Replace
      data: 'Tools'.
"    (w getProperty: (Atom prop: #WmName on: s)
       type: (Atom prop: #String on: s)
       longOffset: 0
       longLength: 256 
       delete: false) inspect."

    w mapWindow.
    s nextEvent.		"ignore expose event"
    self setupHandlers.
    self setupContents.
    self draw.
!

setupHandlers
    w registerFor: ButtonPressEvent
      when: [ :e | e event == w id ]
      do: [ :e | 
		 e detail == 3
		     ifTrue: [ ExitHandlerLoop _ true ].
		 e detail == 1 
		     ifTrue: [ self selectPiece: e ].
		     ].
    w registerFor: ExposeEvent
      when: [ :e | e window == w id ]
      do: [ :dummy | self draw ]
!

setupContents
    | y |
    contents _ OrderedCollection new.
    y _ YStart.
    #(1 2 4 8) do:
	[ :width | contents addLast: 
		       (Lego legoAt: 20@y
			     width: width height: 1 
			     color: 'white' on: w).
		   y _ y + Gap ].
    #(1 2 4 8) do:
	[ :width | contents addLast: 
		       (Lego smallLegoAt: 20@y
			     width: width height: 1 
			     color: 'white' on: w).
		   y _ y + Gap ].
!

unhighlight: pieceIndex
   pieceIndex notNil
       ifTrue: [ (contents at: pieceIndex) drawColor: 'white' ].
   pieceIndex notNil
       ifTrue: [ (contents at: pieceIndex) drawColor: 'white' ].
!

highlight: pieceIndex
   pieceIndex notNil
       ifTrue: [ (contents at: pieceIndex) drawColor: 'lightblue' ].
!

selectPiece: e
    | y piece |
    "5 here is a hack -- should be height of stub"
    y _ ((e eventY - YStart + 5) // Gap)  + 1.
    y ~= prevPiece
	ifTrue: [ self unhighlight: prevPiece.
		  self highlight: y.
		  prevPiece _ y. 
		  piece _ contents at: y.
		  drawingInst width: piece width.
		  drawingInst size: piece size.
		  "### do other selection stuff"
		  
		  ].
!

draw
    contents do: [ :piece | piece draw ]
! !



Smalltalk at: #s put: nil!

"Smalltalk setTraceFlag: -1 to: true!"

| s inst |
    s notNil ifTrue: [s close].
    EventHandlers _ Dictionary new.
    s _ Server connectToDisplay.
    inst _ DemoLego createMainWindowOn: s.
    Pieces createWindowOn: s onInst: inst.
    s processEvents.
    s close

!
