(* Copyright (C) 1992, Digital Equipment Corporation                         *)
(* All rights reserved.                                                      *)
(* See the file COPYRIGHT for a full description.                            *)
(*                                                                           *)
(* Last modified on Tue Nov 23 16:38:48 PST 1993 by kalsow                   *)
(*      modified on Thu May  6 16:25:33 PDT 1993 by muller                   *)

INTERFACE DragonInt;

IMPORT Word;

(* Internal computations of Dragon require the use of bignums.
   Also, the fraction of a floating-point number can occupy more bits
   than there are in a Word.T.  We use an ad-hoc implementation of
   bignums, because only few of the arithmetic operations are needed,
   and only positive numbers are needed.

   A DragonInt.T b represent the value:

      sigma (0, b.s - 1,  b.w[i] * 2 ^ Used * i)

   that is, b.w[0] contains the Used least significant bits, 
   b.w[1] the next Used significant bits and so on. 

   The most significant word of a DragonInt.T must be non-null.

   The value Used is chosen to facilitate the implementation of
   the arithmetic operations needed by the Dragon.F routine:
   intermediate computations on individual Word.T of a DragonInt.T won't 
   overflow. *)
   

TYPE 
  WA = REF ARRAY OF Word.T;
  T = RECORD
    s: INTEGER := 0;
    w: WA := NIL;
  END;

CONST 
  Used = 28;
  Reserved = Word.Size - Used;
  ButTop = Word.RightShift (Word.Not (0), Reserved);

PROCEDURE FixSize (VAR a: T);
  (* computes a.s from a.w, by discarding the null most significant words *)
PROCEDURE copy (READONLY a: T): T;
  (* returns a copy of a *)
PROCEDURE add (READONLY a,b: T): T;
  (* returns the sum of a and b *)
PROCEDURE diff (READONLY a,b: T): T;
  (* returns a - b; b is assumes a >= b *)
PROCEDURE compare (READONLY a, b: T): [-1..1];
  (* returns -1, 0, or 1 when a < b, a = b, or a > b *)
PROCEDURE max (READONLY a, b: T): T;
  (* returns the biggest of a and b *)
PROCEDURE shiftb (READONLY a: T; n: INTEGER): T;
  (* returns a shifted n bits, to the left for n>0, to the right for n<0 *)
PROCEDURE times2 (READONLY a: T): T;
  (* returns 2 * a *)
PROCEDURE timesBinPlace (VAR a: T): T;
  (* modifies a to be 10 * a and returns it *)
PROCEDURE divideB (READONLY a: T): T;
  (* returns a / 10 *)
PROCEDURE divmod (READONLY a, b : T; VAR d: INTEGER): T;
  (* assume that a = d * b + res, which d a number between 0 and 9;
     computes d and returns res. *)

VAR
  Zero, One: T;  (* useful constants *)

END DragonInt.
