(* Copyright (C) 1992, Digital Equipment Corporation                         *)
(* All rights reserved.                                                      *)
(* See the file COPYRIGHT for a full description.                            *)
(*                                                                           *)
(* Last modified on Wed Mar  9 09:21:51 PST 1994 by kalsow                   *)
(*      modified on Mon Mar 16 18:10:15 PST 1992 by muller                   *)

UNSAFE MODULE RTSignal;

IMPORT RTMisc, RTProcess, Usignal, Uprocess;
FROM Ctypes IMPORT int;

TYPE
  SigInfo = UNTRACED REF Usignal.struct_sigcontext;

VAR
  DefaultHandler: Usignal.SignalHandler;
  IgnoreSignal  : Usignal.SignalHandler;

PROCEDURE InstallHandlers () =
  BEGIN
    DefaultHandler := LOOPHOLE (0, Usignal.SignalHandler);
    IgnoreSignal   := LOOPHOLE (1, Usignal.SignalHandler);

    SetHandler (Usignal.SIGHUP,  Shutdown);
    SetHandler (Usignal.SIGINT,  Shutdown);
    SetHandler (Usignal.SIGQUIT, Quit);
    (*** SetHandler (Usignal.SIGBUS,  BusError); ***)
    SetHandler (Usignal.SIGSEGV, SegV);
    SetHandler (Usignal.SIGPIPE, IgnoreSignal);
    SetHandler (Usignal.SIGTERM, Shutdown);
  END InstallHandlers;

PROCEDURE SetHandler (sig: int;  handler: Usignal.SignalHandler) =
  (* Note: we use the LOOPHOLE to prevent the runtime check for
     nested procedure.  The runtime check crashes when
     handler = IgnoreSignal = 1. *)
  VAR new, old: Usignal.struct_sigvec;
  BEGIN
    new.sv_handler := LOOPHOLE (handler, Usignal.SignalHandler);
    new.sv_mask    := Usignal.empty_sv_mask;
    new.sv_flags   := 0;
    EVAL Usignal.sigvec (sig, new, old);
  END SetHandler;

PROCEDURE Shutdown (sig: int; <*UNUSED*> code: int; <*UNUSED*> scp: SigInfo) =
  BEGIN
    RTProcess.InvokeExitors ();                   (* flush stdio... *)
    SetHandler (sig, DefaultHandler);             (* restore default handler *)
    EVAL Usignal.kill (Uprocess.getpid (), sig);  (* and resend the signal *)
  END Shutdown;

PROCEDURE Quit (<*UNUSED*> sig, code: int; scp: SigInfo) =
  VAR pc := 0;
  BEGIN
    IF (scp # NIL) THEN pc := scp.sc_pc END;
    RTMisc.FatalErrorPC (pc, "aborted");
  END Quit;

PROCEDURE SegV (<*UNUSED*> sig, code: int; scp: SigInfo) =
  VAR pc := 0;
  BEGIN
    IF (scp # NIL) THEN pc := scp.sc_pc END;
    RTMisc.FatalErrorPC (pc,
      "Segmentation violation - possible attempt to dereference NIL");
  END SegV;

(*************************************************************
PROCEDURE BusError (<*UNUSED*> sig, code: int; scp: SigInfo) =
  VAR pc := 0;  bad := 0;
  BEGIN
    IF (scp # NIL) THEN  pc  := scp.sc_pc;   (*bad := scp.sc_badvaddr;*)  END;
    IF    (bad = -1) THEN CRE (pc, "narrow, assert, return, or case fault");
    ELSIF (bad = -2) THEN CRE (pc, "range fault");
    ELSIF (bad = -3) THEN CRE (pc, "open array subscript fault");
    ELSIF (bad = -4) THEN CRE (pc, "range fault");
    ELSIF (bad = -5) THEN CRE (pc, "range fault");
    ELSIF (bad = -6) THEN CRE (pc, "narrow fault (unequal array sizes)");
    ELSIF (bad = -7) THEN CRE (pc, "attempt to dereference NIL");
    ELSE RTMisc.FatalErrorPC (pc,"Bus error - possible thread stack overflow");
    END;
  END BusError;

PROCEDURE CRE (pc: INTEGER;  msg: TEXT) =
  BEGIN
    RTMisc.FatalErrorPC (pc, "checked runtime error: ", msg);
  END CRE;
************************************************************)

BEGIN
END RTSignal.
