(* Copyright (C) 1990, Digital Equipment Corporation.         *)
(* All rights reserved.                                       *)
(* See the file COPYRIGHT for a full description.             *)
(*                                                            *)
(* Last modified on Mon Dec  6 16:15:02 PST 1993 by kalsow    *)
(*      modified on Wed Dec 23 15:35:41 PST 1992 by jdd       *)
(*      modified on Thu Nov 12 14:06:13 PST 1992 by muller    *)

INTERFACE RTThread;

IMPORT Csetjmp, Usignal;

CONST
  FramePadBottom = 2; (* additional words before old stack pointer to copy *)
  FramePadTop = 0;     (* additional words beyond new stack pointer to copy *)

(*--------------------------------------------------------- machine state ---*)

TYPE
  State = Csetjmp.jmp_buf;
  (* The machine state is saved in a State.  This type is really opaque
     to the client, i.e. it does not need to be an array. *)

<*EXTERNAL "_setjmp" *> PROCEDURE Save (VAR s: State): INTEGER;

<*EXTERNAL "RTThread__Transfer"*>
PROCEDURE Transfer (VAR from, to: State);
  (* Records the current machine state in "from" and sets
     the machine state to that contained in "to". *)

PROCEDURE SP (READONLY s: State): ADDRESS;
  (* Returns the stack pointer associated with s *)

(*-------------------------------------------------- modifying the models ---*)

(* When a new thread is forked, the contents of its stack and its state
   are built from models.  The only difference with the model is the location
   of the stack.  The two procedures below are used to modify the values of 
   the model to point to the correct state.  offset is the difference 
   between the actual stack location and the model stack location *)

PROCEDURE UpdateStateForNewSP (VAR s: State; offset: INTEGER);
  (* update the state *)

PROCEDURE UpdateFrameForNewSP (a: ADDRESS; offset: INTEGER);
  (* update the stack frame.  a is the correct value for the SP 
     on the stack that contains that frame. *)

(*------------------------------------------------------- flushing caches ---*)

(* On some machines, the stack is cached; for example, on SPARC, the stack
   is partially cached in the registers. FlushStackCache is use to make
   sure that the contents of the stack is up to date. *)

PROCEDURE FlushStackCache ();

(*------------------------------------ manipulating the SIGVTALRM handler ---*)

PROCEDURE setup_sigvtalrm (handler: Usignal.SignalHandler);
PROCEDURE allow_sigvtalrm ();
PROCEDURE disallow_sigvtalrm ();

(*---------------------------------------------------- exception delivery ---*)

<*EXTERNAL "RTThread__handlerStack" *>
VAR handlerStack: ADDRESS;
(* linked list of exception frames. *)

END RTThread.
