(* Copyright (C) 1990, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)
(* Last modified on Fri Mar 25 10:33:48 PST 1994 by kalsow     *)
(*      modified on Sat Jun 27 22:22:30 PDT 1992 by muller     *)

UNSAFE MODULE RTProcess;

IMPORT RTHeapRep, RTException, RTIO, RTOS;

(*-------------------------------- program startup/shutdown -----------------*)

TYPE
  Exitor = BRANDED "RTProcess.Exitor" REF RECORD
             proc: PROCEDURE ();
             next: Exitor;
           END;

VAR
  exitors: Exitor := NIL;

PROCEDURE RegisterExitor (p: PROCEDURE ()) =
  BEGIN
    exitors := NEW (Exitor, proc := p, next := exitors);
  END RegisterExitor;

PROCEDURE InvokeExitors () =
  VAR tmp: Exitor;
  BEGIN
    (* run the registered "exit" routines *)
    WHILE exitors # NIL DO
      (* to ensure progress, remove an element from the list before
         invoking it *)
      tmp := exitors;
      exitors := exitors.next;
      tmp.proc ();
    END;
  END InvokeExitors;

PROCEDURE Exit (n: INTEGER) =
  BEGIN
    InvokeExitors ();
    RTOS.Exit (n);
  END Exit;

PROCEDURE Crash (msg: TEXT) =
  BEGIN
    IF (msg # NIL) THEN
      RTIO.PutText (RTIO.stderr, "\n*** ");
      RTIO.PutText (RTIO.stderr, msg);
      RTIO.PutText (RTIO.stderr, "\n");
    END;
    RTException.DumpStack ();
    RTIO.Flush (RTIO.stderr);

    (* run the registered "exit" routines *)
    InvokeExitors ();

    (* crash *)
    EVAL RTHeapRep.Crash ();
    RTOS.Crash ();
  END Crash;

BEGIN
END RTProcess.



