(* Copyright (C) 1989, Digital Equipment Corporation *)
(* All rights reserved. *)
(* See the file COPYRIGHT for a full description. *)
(* Last modified on Sat Dec  5 11:54:29 PST 1992 by mcjones *)

MODULE FmtTime;

IMPORT Date, Fmt, Time;

CONST
  S = " ";
  C = ":";

PROCEDURE DateLong(READONLY d: Date.T): TEXT =
  BEGIN
    WITH
      DDD = WeekDay[d.weekDay],
      MMM = Month[d.month],
      dd = Fmt.Pad(Fmt.Int(d.day), 2),
      hh = Fmt.Pad(Fmt.Int(d.hour), 2, '0'),
      mm = Fmt.Pad(Fmt.Int(d.minute), 2, '0'),
      ss = Fmt.Pad(Fmt.Int(d.second), 2, '0'),
      zz = d.zone,
      yyyy = Fmt.Pad(Fmt.Int(d.year), 4)
    DO
      RETURN
        DDD & S & MMM & S & dd & S & hh & C & mm & C & ss & S & zz & S & yyyy
    END
  END DateLong;

PROCEDURE DateShort(READONLY 
d: Date.T): TEXT =
  BEGIN
    WITH
      MMM = Month[d.month],
      dd = Fmt.Pad(Fmt.Int(d.day), 2),
      hh = Fmt.Pad(Fmt.Int(d.hour), 2, '0'),
      mm = Fmt.Pad(Fmt.Int(d.minute), 2, '0')
    DO
      RETURN MMM & S & dd & S & hh & C & mm
    END
  END DateShort;

PROCEDURE Long(t: Time.T; z: Date.TimeZone := NIL): TEXT =
  BEGIN
    RETURN DateLong(Date.FromTime(t, z))
  END Long;

PROCEDURE Short(t: Time.T; z: Date.TimeZone := NIL): TEXT =
  BEGIN
    RETURN DateShort(Date.FromTime(t, z))
  END Short;

BEGIN
END FmtTime.
