(* Copyright (C) 1993, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)
(* Last modified on Tue Aug 24 08:38:41 PDT 1993 by kalsow     *)
(*      modified on Thu Jul 15 16:23:08 PDT 1993 by swart      *)

UNSAFE MODULE MachineIDPosix EXPORTS MachineID;

IMPORT Unix, Usocket, Unetdb;

FROM Word IMPORT Shift, And, Or;

TYPE
  Byte = BITS 8 FOR [0..255];

(* The following information is commented out in the Unix interface.  This
   is probably system dependent. *)

TYPE
  struct_ifreq = RECORD
    ifr_name  : ARRAY [0 .. 15] OF CHAR;  (* if name, e.g.  "en0" *)
    ifru_addr : Usocket.struct_sockaddr;
  END;

  struct_ifconf = RECORD
    ifc_len  : INTEGER;  (* size of associated buffer *)
    ifcu_req : UNTRACED REF struct_ifreq;
  END;

  struct_ifdevea = RECORD
    ifr_name   : ARRAY [0 .. 15] OF CHAR;  (* if name, e.g.  "en0" *)
    default_pa : ARRAY [0 .. 5] OF Byte;   (* default hardware address *)
    current_pa : ARRAY [0 .. 5] OF Byte;   (* current physical address *)
  END;

CONST
  IFD = Shift(And(BYTESIZE(struct_ifdevea), Unix.IOCPARM_MASK), 16);
  IFC = Shift(And(BYTESIZE(struct_ifconf), Unix.IOCPARM_MASK), 16);

  SIOCGIFCONF   = Or(Unix.SIOCGIFCONF, IFC);   (* Get ifnet ls.*)
  SIOCRPHYSADDR = Or(Unix.SIOCRPHYSADDR, IFD); (* Read phy.  ad.*)

PROCEDURE Get (): T =
  VAR
    hostname: ARRAY [0 .. 33] OF CHAR;
    hostent : Unetdb.struct_hostent_star;
    id      : T;
    list    : struct_ifconf;
    buf     : ARRAY [0 .. 10] OF struct_ifreq;
    req     : struct_ifdevea;
    s       : INTEGER;
  BEGIN
    (* try to find an ethernet hardware address *)

    s := Usocket.socket(Usocket.AF_UNIX, Usocket.SOCK_STREAM,
                        Usocket.PF_UNSPEC);
    IF (s >= 0) THEN
      list.ifc_len := BYTESIZE(buf);
      list.ifcu_req := ADR(buf);
      IF Unix.ioctl(s, SIOCGIFCONF, ADR(list)) # -1 THEN
        FOR i := 0 TO list.ifc_len DIV BYTESIZE(struct_ifreq) - 1 DO
          req.ifr_name := buf[i].ifr_name;
          IF (Unix.ioctl(s, SIOCRPHYSADDR, ADR(req)) # -1)
            AND (req.default_pa[0] # 0 OR req.default_pa[1] # 0) THEN
            id.r := req.default_pa;
            RETURN id;
          END;
        END;
      END;
    END;


    (* try using the machine's internet address *)

    IF Unix.gethostname(ADR(hostname), 32) # 0 THEN Die (); END;
    hostent := Unetdb.gethostbyname(ADR(hostname));
    IF hostent = NIL THEN Die (); END;
    IF hostent.h_length # 4 THEN Die (); END;
    TYPE Ptr = UNTRACED REF ARRAY [0 .. 3] OF Byte;
    VAR p : Ptr := LOOPHOLE (hostent.h_addr_list^, Ptr); BEGIN
      id.r[0] := 0;
      id.r[1] := 0;
      id.r[2] := p[0];
      id.r[3] := p[1];
      id.r[4] := p[2];
      id.r[5] := p[3];
    END;
    RETURN id;
  END Get;


EXCEPTION Failure;

PROCEDURE Die () =
  <*FATAL Failure*>
  BEGIN
    RAISE Failure;
  END Die;

BEGIN
END MachineIDPosix.
